! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!*************************************************************************************************
! omsrclistcomb
!
! Module: omsrclistcomb_mod
!
! Authors:  Christopher Brindle - MSSL
!           Andrew King - MSSL
!           Tim Harries - MSSL
!          
! Description: Combines OSW source list into an observation source list
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History: 
!
! 2013-01-14 vny: fixed the problem of erroneous pass of the RA and DEC data
! to the subroutine DoAstrometryCorrection (the array arguments taken from the structure source
! were not passed properly causing a crash of the program) 
!
!************************************************************************************************
module omsrclistcomb_mod
  use param
  use cal
  use dal
  use Errorhandling
  use sla
  use caltypes 
  use omallocation_mod
  use omconstants_mod
!  use omastromcorrect_mod
  use omutility_mod
  use omgraph_mod
!  use fluxfit_mod
  use omsourceflags_mod
  use OmSourceParam_mod    
  use omimagealign_mod
  use getfilterid_mod  
  
  use omcoinccorr_mod
  implicit none 

  private 

  public omsrclistcomb

  !*****************************
  ! Structure to define a color
  !*****************************
  type ColourT
     integer :: mag1
     integer :: mag2
     character(20) label, label1
     character(20) elabel
     character(20) flabel
  end type ColourT

!******************************************************************
!
! The OM filters
!
!******************************************************************
integer(int32), private :: filt(7) =  (/ UVW2, UVM2, UVW1, U, B, V, WHITE /)

integer(int16), private,  pointer     :: qflagTable(:, :) => null()       ! Quality flag for each filter
integer(int8), private,  pointer     :: cflagTable(:, :) => null()        ! Confusion flag table
integer(int8), private,  pointer     :: eflagTable(:, :) => null()        ! Extension flag table

real(single), private, pointer       :: mag_table(:, :) => null()         ! table of magnitudesr
real(single), private, pointer       :: magerr_table(:, :) => null()      ! table of magnitude errors
real(single), private, pointer       :: corrRate_table(:, :) => null()    ! table of corrected count_rates
real(single), private, pointer       :: rawRate_table(:, :) => null()     ! table of raw count_rates


!*************************************************************************************
!
! January 2008- For Sky-images need the raw background-count-rate
!
!*************************************************************************************/

real(single), private, pointer       :: rawBackgroundRate_table(:, :) => null()     ! table of raw background count_rates



real(single), private, pointer       :: corrRateErr_table(:, :) => null() ! table of magnitude error
real(single), private, pointer       :: majorTable(:,:) => null()         ! table of major axes
real(single), private, pointer       :: minorTable(:,:) => null()         ! table of major axes
real(single), private, pointer       :: paTable(:,:) => null()            ! table of major axes
real(single), private, pointer       :: significanceTable(:, :) => null() ! table of magnitudes
real(double), private :: raz, decz
real(single), private :: pascx
real(single), private :: nsigma
logical, private, pointer        :: magexist(:,:) => null()               ! does a magnitude exist?


!type(SourceData), pointer :: omFitData(:) => null() ! Stores the data needed for the template-spectrum fitting
 
real(single) :: tolerance = 2.5 ! was 2.5
integer(int32), parameter, private               :: int32NullValue = -1      
integer(int8), parameter, private                :: int8NullValue = 127_int8      
integer(int16), parameter, private               :: int16NullValue = 32767_int16      



real(double), parameter :: arcSecsToRadians = degreesToRadians / 3600.0_double ! To convert from arcsecs to radians

character(10), private :: omDataMode ! The OM data type (IMAGING, FAST, GRISM, MOSAICED)

contains 

!*************************************************************************************
! The main program
!
!*************************************************************************************
subroutine omsrclistcomb

    integer, parameter          :: nFilt = 7             ! no of filters
    integer, parameter          :: nCol = 15             ! no of colours for instrumental magnitudes
    integer, parameter          :: nCol1 = 8             ! no of colours for standard magnitudes
    type(ColourT)               :: instrumentalColour(nCol)
    type(ColourT)               :: standardColour(nCol)

    type(DataSetT)              :: oswDataSet            ! handles
    type(DataSetT)              :: outDataSet            ! handle for output dataset
    type(TableT)                :: tableSet
  
    character(10), pointer      :: filter(:) => null()             ! filters for osw source lists
    

    character(1024), pointer    :: oswFile(:) => null()             ! filenames
    character(50)               :: label                  ! column label
    integer                     :: nStart                 ! start no of sources
    integer                     :: nIndep                 ! number of independent sources
    integer                     :: nosw                   ! no of osw lists
    integer                     :: i, j, k, n             ! counters
    integer                     :: nSources               ! total no of sources
    integer                     :: nSourcesModified       ! mofofied no of sources
    integer(int32), pointer     :: nrows(:)=>null()       ! no of rows in each osw source list
 
    ! Introducing mag limit
    real(single), pointer       :: maglimit(:) => null() ! Magnitude limit for each filter   
    real(single) :: magLimitValue
    type(AttributeT) :: magLimitAttribute
    integer :: ii
    !
    real(single), pointer       :: magColour(:, :) => null()   
    real(single), pointer       :: magColourError(:, :) => null()   
    real(single), pointer       :: standardMagnitude(:, :) => null()  
    real(single), pointer       :: standardMagnitudeError(:, :) => null()
    real(single), pointer       :: stdMagnitudeStore(:, :, :) => null()  
    real(single), pointer       :: stdMagnitudeErrorStore(:, :, :) => null()
   
    real(single), pointer       :: magnitude(:, :) => null()   
    real(single), pointer       :: magnitudeError(:, :) => null()   

    real(double), pointer       :: galong(:) => null()      ! array for galactic longitudes
    real(double), pointer       :: galat(:) => null()       ! array for galactic latitudes    
    real(double)                :: stdColour
    real(double)                :: stdMagnitude1
    real(double)                :: stdMagnitude2
    real(double)                :: stdColourError
    real(double)                :: stdMagnitude1Error
    real(double)                :: stdMagnitude2Error
      
    integer                     :: noColours
    character(100)              :: colourString(15), stdString(15)
    character(100)              :: colourErrorString(15)
    character(200)              :: filterString, stdString1, stdColourString
    real(single), pointer       :: stdColourMagnitude(: , :) => null()
    real(single), pointer       :: stdColourMagnitudeError(: , :) => null()
    integer(int32), pointer     :: RASortIndex(:) => null()
    integer                     :: iColumn, jColumn      
    integer                     :: noFilters, noFluxFilters
    integer                     :: filterIndex(20), fluxFilterIndex(20)    
    integer                     :: allocErr, position
    integer(int32), pointer     :: nCount(:, :) => null()           

    logical, dimension(1 : nCol) :: stdMagExist          ! Does a standard magnitude exist?
    real(single) :: expos
    integer(int8) :: tol    
    character(len=5)  :: filtid(nFilt) = &
               (/'UVW2 ','UVM2 ','UVW1 ','U    ','B    ','V    ', 'WHITE'/) 
 
    integer :: filtIndex0
    !integer :: filtIndex(1:nfilt)
    integer(int32), pointer :: filtIndex(:) => null() 
    real(single) :: averMaglimit(1:nfilt)
    real(single) :: naverMagLimit(1:nfilt)

    
    character(200) :: string
    type(sourceParameters), pointer :: source(:) => null()
    character(100) :: comment  
    type(sourceParameters), pointer :: newsource(:) => null() 



    !**************************************************************************
    ! The following are for attempting to align the windows
    !*************************************************************************

    integer :: noWindows ! , noExposures 
!    type(ExposureList), pointer :: expList(:) => null()
    character(4), pointer :: exposureId(:) => null()
    type(WindowList), pointer :: winList(:) => null()

   
    !********************************************************************************
    ! 2013-01-14 vny: passing the RA and DEC arrays to the routine DoAstrometryCorrection
    ! (instead of passing source%rightAscension and source%declination)
    ! because the data inside of the routine was wrong
    !*******************************************************************************
    real(double), pointer :: raCorr(:) => null()
    real(double), pointer :: decCorr(:) => null()
    real(single), pointer :: poserrCorr(:)=> null()

!*********************************************************************
! The following is for testing purposes onle- should be blanked out
!**********************************************************************
!open(unit=10, status="UNKNOWN", file="OMSRCLISTCOMB.DAT")

    radiansToArcSeconds = radiansToDegrees * 3600.0_double


    !************************************************
    ! Initialise the filter prority and colour 
    ! tables for the instrumental magnitudes
    !************************************************
    call SetUpInstrumentalColourTable(instrumentalColour)


    !************************************************
    ! Initialise the filter prority and colour 
    ! tables for the standard magnitudes
    !************************************************
    call SetUpStandardColourTable(standardColour)

    !************************************************
    ! Get the number of osw files and their names.
    !************************************************

    nosw = parameterCount("sourcelistsets")

    if (nosw .eq. 0) then

        call warning("Can't continue", "No source-list files to process")
        stop
    end if    

 
    oswFile => AllocateChar(1024, nosw, "oswFile")

    nsigma = realParameter("nsigma")

    nrows => AllocateInt32(1, nosw, "nrows")   ! The number of rows in each source-list file
    maglimit => AllocateSingle(1, nosw, "maglimit")   ! The number of rows in each source-list file
    filtIndex => AllocateInt32(1, nosw, "filtIndex")  
    !write(*,*) "L245 maglimit is allocated, nosw=", nosw

    filter => AllocateChar(10, nosw, "filter") ! The filter in each source-list file
    
    exposureId => AllocateChar(4, nosw, "exposureid") ! The exposue id in each source-list file

    nSources = 0 
 
    !********************************************************************************************
    !
    ! Determine total number of sources and number of source-list files with at least one source
    !
    !********************************************************************************************

    nosw = 0
    !write(*,*) "L261 parameterCount=", parameterCount("sourcelistsets")
    !write(*,*) "L262 size(oswfile)=", size(oswfile)

    do i=1, nfilt
       averMagLimit(i)=0.0
       naverMagLimit(i)=0.0
    enddo
    do i = 1 , parameterCount("sourcelistsets")
                
        oswfile(i) = stringParameter("sourcelistsets", i - 1)
       
        oswDataSet = DataSet(oswFile(i),read)
      
        tableSet = table(oswDataSet, 0)
        nrows(i) = numberOfRows(tableSet)
    
        if (nrows(i) .gt. 0) then
            nSources = nSources + nrows(i)
            nosw = nosw + 1
            oswfile(nosw) = oswfile(i)
            nrows(nosw) = nrows(i)
            filter(nosw) = trim(stringAttribute(oswDataSet, "FILTER"))          


            filtIndex0=0
            if (lge(trim(filter(nosw)),"UVW2") .and. lle(trim(filter(nosw)),"UVW2") ) then
               filtIndex0=1
               !write(*,*) "L276 filter is iqual to UVW2, filtIndex0=", filtIndex0
            else
               if ( lge(trim(filter(nosw)),"UVM2") .and. lle(trim(filter(nosw)),"UVM2") ) then
                  filtIndex0=2
                  !write(*,*) "L280 filter is iqual to UVM2 filtIndex0=", filtIndex0
               else
                  if ( lge(trim(filter(nosw)),"UVW1") .and. lle(trim(filter(nosw)),"UVW1") ) then
                     filtIndex0=3
                     !write(*,*) "L284 filter is iqual to UVW1 filtIndex0=", filtIndex0
                  else
                     if ( lge(trim(filter(nosw)),"U") .and. lle(trim(filter(nosw)),"U") ) then
                        filtIndex0=4
                        !write(*,*) "L288 filter is iqual to U filtIndex0=", filtIndex0
                     else
                        if ( lge(trim(filter(nosw)),"B") .and. lle(trim(filter(nosw)),"B") ) then
                           filtIndex0=5
                           !write(*,*) "L293 filter is iqual to B filtIndex0=", filtIndex0
                        else
                           if ( lge(trim(filter(nosw)),"V") .and. lle(trim(filter(nosw)),"V") ) then
                              filtIndex0=6
                              !write(*,*) "L296 filter is iqual to V filtIndex0=", filtIndex0
                           else
                              if ( lge(trim(filter(nosw)),"WHITE") .and. lle(trim(filter(nosw)),"WHITE") ) then
                                 filtIndex0=7
                                 !write(*,*) "L300 filter is iqual to WHITE filtIndex0=", filtIndex0
                              endif
                           endif
                        endif
                     endif
                  endif
               endif
            endif
            !write(*,*) "L313 i=", i, " nosw=", nosw, " filtIndex0=", " size(filtIndex)=", &
            !     size(filtIndex)
            if (filtIndex0 .gt. 0) then
               filtIndex(nosw)=filtIndex0
            endif

            magLimitValue=0.0
            if( hasAttribute( oswDataSet, "MAGLIMIT" ) ) then               
               magLimitAttribute=attribute( oswDataSet, "MAGLIMIT" )
               magLimitValue=real32Attribute(magLimitAttribute)
               maglimit(nosw)=magLimitValue
               if (filtIndex0 .gt. 0 .and. filtIndex0 .lt. 8) then
                  averMagLimit(filtIndex0)=averMagLimit(filtIndex0)+magLimitValue
                  naverMagLimit(filtIndex0)=naverMagLimit(filtIndex0)+1.0
               endif
            endif
            
            !write(*, *) "L330 i=", i, " nosw=", nosw, & !" oswfile=", trim(oswfile(nosw)), &
            !     " filter=", filter(nosw), " nrows=", nrows(nosw), " magLimitValue=", magLimitValue, &
            !     " maglimit=", maglimit(nosw), " filtIndex=", filtIndex(nosw)


        end if
        call release(tableSet)
        call release(oswDataSet)

    end do

    do i=1,nfilt
       if (naverMagLimit(i) .gt. 1.0) then
          averMagLimit(i)=averMagLimit(i)/naverMagLimit(i)
       endif
       !write(*,*) "L356 i=", i, " filtId=", filtid(i), " averMagLimit=", averMagLimit(i), &
       !     " n=", naverMagLimit(i)
    enddo




    !*************************************
    ! Check there is at least 1 source
    !*************************************

    if (nSources .eq. 0) then
       
        call warning("Can't continue", "No sources to process")


        !******************************************
        ! create the output observation source list
        !******************************************
        nosw = 1
        oswfile(1) = stringParameter("sourcelistsets", 0)
    
        call CreateObsSourceList(nOsw, oswFile, outDataSet, raz, decz, expos)    
          
        write(*,*) "L389 raz=", raz, " decz=", decz

        !**************************************************
        ! Create the observation source list
        ! which has nIndep rows
        !**************************************************
        tableSet = addTable(outDataSet, "SRCLIST", 0)
        call release(outdataSet)

        !***********************************************************
        ! Can't do anything else, so stop here
        !**********************************************************
        stop
    end if    







    !******************************************
    ! Allocate memory for the arrays to hold
    ! the input data.
    !*****************************************
 

    !**********************************************************************
    ! The array of structs to hold data from individual source-list files
    !**********************************************************************
    allocate(newSource(1 : nSources)) 
 

    !********************************************************
    ! Read in the data from each osw source list into the 
    ! appropriate array- appending the new data to the end of
    ! that array.
    !********************************************************
     

    write(errstr,'(a)') "Loading data from source-list files"
    call message(AppMsg, Verbose)


    nStart = 0






    k = 0 ! For mosaiced source-lists only want to keep the faint sources- so that the number of sources will change
    nSources = 0 

    do i = 1 , nosw

    !    write(*, *) i, ") line 334 nrows=", nrows(i)
        !************************************************************************
        !
        ! Load the data from this source-list file and append to the source-list
        !
        !***********************************************************************   
        call LoadSourceData(oswFile(i), newSource, nStart, nRows(i))

    !    write(*, *) i, ") line 342 nrows=", nrows(i)
  
        if(nRows(i) > 0) then
            k = k + 1
            oswfile(k) = oswfile(i)
            nrows(k) = nrows(i)
            filter(k) = filter(i)          
    !        write(*, *) i, ") line 349 k=", k, " nrows=", nrows(k), " nsources=", nsources, " nstart=", nStart
  
            newSource(nStart + 1 : nstart + nrows(i))%imin = k ! imin records the file from where the source originated
     
    !        write(*, *) i, ") line 353 k=", k, " nrows=", nrows(k), " nsources=", nsources
  

            nStart = nStart + nrows(i)  
            nRows(k) = nrows(i)
            nSources = nSources + nrows(i)
    !        write(*, *) i, ") line 355 k=", k, " nrows=", nrows(k), " nsources=", nsources
  

        end if
    end do
    nosw = k  ! The new number of useable input files


    !*************************************
    ! Check there is at least 1 source
    !*************************************






    if (nSources .eq. 0) then
        deallocate(newSource)
        nullify(newSource)
        call warning("Can't continue", "No sources to process")


        !******************************************
        ! create the output observation source list
        !******************************************
        nosw = 1
        oswfile(1) = stringParameter("sourcelistsets", 0)
    
        call CreateObsSourceList(nOsw, oswFile, outDataSet, raz, decz, expos)    
          

        !write(*,*) "L502 raz=", raz, " decz=", decz
        !**************************************************
        ! Create the observation source list
        ! which has nIndep rows
        !**************************************************
        tableSet = addTable(outDataSet, "SRCLIST", 0)
        call release(outdataSet)

        !***********************************************************
        ! Can't do anything else, so stop here
        !**********************************************************
        stop
    end if    





     
    !******************************************
    ! create the output observation source list
    !******************************************
    call CreateObsSourceList(nOsw, oswFile, outDataSet, raz, decz, expos)    
    
    !write(*,*) "L526 raz=", raz, " decz=", decz
    raz = raz * degreesToRadians
    decz = decz * degreesToRadians

    allocate(source(1 : nSources))
    !
    ! 2013-01-14 vny
    ! 2013-05-14 vny allocating raCorr and decCorr near the subroutine 
    ! DoAstrometryCorrection
    !allocate(raCorr(1 : nSources)) 
    !allocate(decCorr(1 : nSources))
    !allocate(poserrCorr(1 : nSources))
    !
    source(1 : nSources) = newSource(1 : nSources) 

    deallocate(newSource)
    nullify(newSource)

    
    where (source(1 : nsources)%poserr .gt. 5.0)
        source(1 : nsources)%poserr = 5.0
    end where

    where (source%magnitude .gt. 100) 
        source%magnitude = NullValue
        source%magnitudeerr = NullValue
        source%corrcountRate = NullValue
        source%corrCountRateErr = NullValue
       
    end where
  
    !***************************************
    ! Do a few conversions
    !***************************************
    source%rightascension = source%rightascension * degreesToRadians
    source%declination = source%declination * degreesToRadians
    !
    !2013-01-14
    !raCorr=source%rightascension
    !decCorr=source%declination
    !poserrCorr=source%poserr
    !write(*,*) "L564 -------------"
    !do i =1, 3
    !   write(*,*) "L566 ra=", source(i)%rightascension!, " raCorr=", raCorr(i)
    !enddo

    !
    if (booleanParameter("alignaxes")) then  
        call getWindowList(nosw, oswfile, filter, noWindows, winList)
        if(noWindows .gt. 0) then
            if( omDataMode .eq. "MOSAICED" ) then
                 tol = 10
            else
                 tol = 2
            end if

            call AlignWindows(tol, source, nosw, nrows, winList, filtId)
            if (associated(winList)) deallocate(winList)
        end if
    end if

    !************************************************************************************
    ! Create the merged-source-list file
    !
    !************************************************************************************
    call CreateMergedSourceList(nSources, source, nosw, oswfile,      &
                                filter, nrows, nfilt, filtId,         &
                                nindep, newSource)



!close(10)

      
    !**************************************************
    ! Create the observation source list
    ! which has nIndep rows
    !**************************************************
    tableSet = addTable(outDataSet, "SRCLIST", nIndep)

    RASortIndex => AllocateInt32(1, nIndep, "RASortIndex")
 
    do i = 1, nIndep 
        RASortIndex(i) = i
    end do



    call AddInt32IndexColumn(tableSet, "SRCID", "", "Source identification", nIndep, RASortIndex, RASortIndex)
 
    !*****************************************************
    ! Find the order of increasing  RA- store in RASortIndex
    ! for adding data to source list table in  
    ! increasing RA order (ra is not sorted)
    !****************************************************
    call IndexSort(newSource%rightAscension, RASortIndex, nIndep)


    !**********************************************************
    ! Add a source ID column to each of the osw source lists
    ! telling us which source went to which entry in the
    ! observation source list 
    !**********************************************************
    call AddSourceIds(nosw, oswfile, nrows, nindep, source, raSortIndex)
    deallocate(source)
    source => newSource
    !****************************************************
    ! Compute some statistics on astrometry
    !****************************************************
    
!    call CheckIndividualSourceLists( source, raSortIndex )



  

 

    call AddReal64IndexColumn(tableSet, "RA", "degrees", "Averaged RAs of identical sources", &
                        nIndep, source%rightascension * radiansToDegrees, RASortIndex, nullValue)

    call AddReal64IndexColumn(tableSet, "DEC", "degrees", " Averaged decs of identical sources", &
                    nIndep, source%declination * radiansToDegrees, RASortIndex, nullValue)
   

    call AddReal32IndexColumn(tableSet, "POSERR", "arcsec", " ", nIndep, source%poserr, &
     RASortIndex, nullValue)
 

 
    !**************************************
    ! Find out what filters have magnitudes
    !************************************** 
    noFilters = 0
    noFluxFilters = 0
    filterString = ""
    
    call addcomment(tableSet, "The table consists of data taken with the following filters")

  

    do i = 1 , nFilt
   
        if (ANY(magexist(1 : nIndep, i : i)))  then
            
            noFluxFilters = noFluxFilters + 1
            noFilters = noFilters + 1
            filterIndex(noFilters) = i
            write(comment, *) noFilters, ") ", trim(filtid(i))
            call addcomment(tableset, trim(comment))
        
        
        

            if (noFluxFilters .gt. 1) then
                fluxFilterIndex(noFluxFilters) = i
                label = trim(filterString)
                filterString = trim(label) // ","  // trim(filtid(i))
            else
                filterString = trim(filtid(i))
            end if
          else
        
            write(errstr, '(a, a)' ) "No sources found in filter ", trim(filtid(i))
            call message(AppMsg, Verbose)
        end if
    end do
 
    !******************************************************
    ! Add the significance arrays
    !****************************************************** 
!    if( omDataMode .eq. "MOSAICED" ) then
!        call setSignificanceTable( significanceTable, rawBackgroundRate_Table, rawRate_table)=         
!    end if

    do i = 1, noFilters

        call AddReal32IndexColumn(tableSet, trim( filtid(filterIndex(i) )) // "_SIGNIFICANCE", &
                  "", "Significance of detection", &
                  nIndep, significanceTable(:, filterIndex(i) ), RASortIndex, nullValue)

    end do
 

 
 
    deallocate(significanceTable, stat=allocErr)
    call CheckDeallocation(allocErr, "significance")



    !***********************************************************************************
    !
    ! If the SWS files have come from a sky-image add the raw rates
    !
    !
    !***********************************************************************************
    if( omDataMode .eq. "MOSAICED" ) then
 

        do i = 1, noFilters

  !           call AddReal32IndexColumn(tableSet, trim( filtid(filterIndex(i) )) // "_RAW_BACKROUND_RATE", &
  !                         "count/sec", "The raw-background count-rates", &
  !                nIndep, rawBackgroundRate_Table(:, filterIndex(i) ), RASortIndex, nullValue)
             k = filterIndex(i)
             do j = 1, nIndep
                
                 if( .not. isNullValue( rawRate_Table(j, k )) .and. rawRate_Table(j, k) >= 0.0) then 
                     rawRate_Table(j, k ) = rawRate_Table(j, k ) + &
                                     rawBackgroundRate_Table(j, k ) 
                 end if
             end do

             call AddReal32IndexColumn(tableSet, trim( filtid(k) ) // "_RAW_RATE", &
                  "count/sec", "The raw count-rates", &
                  nIndep, rawRate_Table(:, k ), RASortIndex, nullValue)


    end do







    end if
    !***********************************************************
    ! Add the mean count-rates
    !**********************************************************

    do i = 1, noFilters

        k = filterIndex(i)
       
        call AddReal32IndexColumn(tableSet, trim( filtid( k )) // "_CORR_RATE", &
                           "count/sec", "The corrcted count-rates", &
                  nIndep, corrRate_Table(:, k ), RASortIndex, nullValue)

        call AddReal32IndexColumn(tableSet, trim( filtid(k )) // "_CORR_RATE_ERR", &
                  "count/sec", "The corrected count-rate errors", &
                  nIndep, corrRateErr_Table(:, k ), RASortIndex, nullValue)


    end do




    !***********************************************************
    ! Add the instrumental magnitudes
    !**********************************************************
    do i = 1, noFilters

        k = filterIndex(i)
    

        call AddReal32IndexColumn(tableSet, trim( filtid(k )) // "_INS_MAG", "mag", "Instrumental magnitudes", &
                  nIndep, mag_Table(:, k ), RASortIndex, nullValue)


        call AddReal32IndexColumn(tableSet, trim( filtid(k )) // "_INS_MAG_ERR", &
         "mag", "Instrumental magnitude errors ", &
                  nIndep, magErr_Table(:, k ), RASortIndex, nullValue)

    end do




    !*********************************************
    ! Add the instrumental colours
    !*********************************************
    magColour => Allocate2DSingle(1, nIndep, 1, ncol, "magColour")
    magColourError => Allocate2DSingle(1, nIndep, 1, ncol, "magColourError")
    magColour = NullValue
    magColourError = NullValue

    n = 0
    string = ""


  
    do i = 1, nCol     ! Loop through all the possible colours
        iColumn = instrumentalColour(i)%mag1
        jColumn = instrumentalColour(i)%mag2
        if (ANY(magexist(1 : nIndep, iColumn : iColumn)) .and.  &
            ANY(magexist(1 : nIndep, jColumn : jColumn))) then
            n = n + 1
            do k = 1, nIndep
                if (mag_table(k, iColumn) .ne. NullValue .and. &
                    mag_table(k, jColumn) .ne. NullValue) then
                    magColour(k, n) = mag_table(k, iColumn) - mag_table(k, jColumn)
                    magColourError(k, n) = AddErrors(dble(magerr_table(k, iColumn)), dble(magerr_table(k, jColumn)))
                end if
            end do 
            if (n .eq. 1) then
                string = trim(instrumentalColour(i)%label1)
            else
                label = string
                string = trim(label) // "," //   trim(instrumentalColour(i)%label1)
                
            end if

            label = trim(instrumentalColour(i)%label1) // "_INS"
            write(label, *) "Column ", n, "=", trim(instrumentalColour(i)%label1)  


            call AddReal32IndexColumn(tableSet, trim(instrumentalColour(i)%label) // "_INS", "mag", "Instrumental colours ", &
                  nIndep, magColour(:, n), RASortIndex, nullValue)

            call AddReal32IndexColumn(tableSet, trim(instrumentalColour(i)%label1) // "_INS_ERR", "mag", "Ins colour errors ", &
                  nIndep, magColourError(:, n), RASortIndex, nullValue)


        end if
    end do




    !*******************************************************************
    ! now we have the colours
    ! we can currently have (U-B) and/or (B-V), but the combinations
    ! can be added straightfowardly by changing the array in SetUpColours
    !********************************************************************
    call CAL_setState(OM , 1, accuracyLevel = ACCURACY_HIGH)



    !****************************************
    ! Allocate memory for arrays holding
    ! the standard colours
    !*************************************** 
    stdColourMagnitude => Allocate2DSingle(1, nIndep, 1, ncol, "stdColourMagnitude")
    stdColourMagnitude = NullValue  

    stdColourMagnitudeError => Allocate2DSingle(1, nIndep, 1, ncol, "stdColourMagnitudeError")
    stdColourMagnitudeError = NullValue  

    !***************************************************
    ! Calculate the standard magnitudes and colours
    ! NB- can only use U, B and V  
    !***************************************************
    standardMagnitude => Allocate2DSingle(1, nIndep, 1, nCol, "standardMagnitude")
    standardMagnitude = NullValue

    standardMagnitudeError => Allocate2DSingle(1, nIndep, 1, nCol, "standardMagnitudeError")
    standardMagnitudeError = NullValue    

    stdMagExist = .false.
    noColours = 0

    !*****************************************************************
    ! Calculate the standard colours and magnitudes in reverse order 
    ! to preserve the priority
    ! Store the different standard magnitudes and errors for each source
    ! that have been computed for each colour combination (then average
    ! them after and calculate the colours  
    !*****************************************************************

    nCount => Allocate2DInt32(1, nCol, 1, nCol, "nCount") 
 
    stdMagnitudeStore => Allocate3DSingle(1, nIndep, 1, nCol, 1, nCol,   "stdMagnitudeStore")
    stdMagnitudeErrorStore => Allocate3DSingle(1, nIndep, 1, nCol, 1, nCol,   "stdMagnitudeErrorStore")
   
    stdMagnitudeStore = NullValue
    stdMagnitudeErrorStore = NullValue


    nCount = 0
    stdString1 = ""
    stdColourString = ""
    noColours = 0  ! The number of instrumental colours

!    call addcomment(tableSet, "STANDARD MAGNITUDES- COLUMN FILTER CODES-")
   
    do i = 3, 1, -1 ! Only first 3 colours are optical

        iColumn = standardColour(i)%mag1
        jColumn = standardColour(i)%mag2

     
        if (ANY(magexist(1 : nIndep, iColumn : iColumn)) .and.  &
            ANY(magexist(1 : nIndep, jColumn : jColumn))) then

            noColours = noColours + 1
            colourString(noColours) = standardColour(i)%label1
            colourErrorString(noColours) = standardColour(i)%elabel
            stdString(noColours) = standardColour(i)%flabel

            if (noColours .eq. 1) then
                stdString1 = trim(standardColour(i)%flabel)
                stdColourString = trim(standardColour(i)%label1)
                    
            else
                label = trim(stdString1)
                stdstring1 = trim(label) // "," //  trim(standardColour(i)%flabel)

                label = trim(stdColourString)
              
                stdColourstring = trim(label) // "," //  trim(standardColour(i)%label1)
            end if




            stdString(noColours) =  trim(standardColour(i)%label1)
            stdMagExist(iColumn) = .true.
            stdMagExist(jColumn) = .true.
            
            do k = 1, nIndep
                if (mag_table(k, iColumn) .ne. NullValue .and. &
                    mag_table(k, jColumn) .ne. NullValue) then

                    !************************************************
                    ! Calculate the standard magnitudes and colour
                    !************************************************
                    call CalculateStandardColour(filt(iColumn), filt(jColumn),        &
                        dble(mag_table(k, iColumn)),  dble(magerr_table(k, iColumn)), &
                        dble(mag_table(k, jColumn)),  dble(magerr_table(k, jColumn)), &
                        stdMagnitude1, stdMagnitude1Error,                            &
                        stdMagnitude2, stdMagnitude2Error,                            &
                        stdColour, stdColourError)
                    !**************************************************************
                    ! Set nCount to 1 to record that this colour was used to
                    ! compute a standard colour
                    !************************************************************** 
                    nCount(iColumn, noColours) = 1
                    nCount(jColumn, noColours) = 1
                
                    stdMagnitudeStore(k, iColumn, noColours) = stdMagnitude1
                    stdMagnitudeErrorStore(k, iColumn, noColours) = stdMagnitude1Error 
                    stdMagnitudeStore(k, jColumn, noColours) = stdMagnitude2
                    stdMagnitudeErrorStore(k, jColumn, noColours) = stdMagnitude2Error
                    stdColourMagnitude(k, noColours) = stdColour
                    stdColourMagnitudeError(k, noColours) = stdColourError
    
                end if





            end do 

        end if
    end do
  
    deallocate(magexist, stat=allocErr)
    call CheckDeallocation(allocErr, "magexist")

   
    magnitude => Allocate2dSingle(1, nIndep, 1, nFilt, "magnitude")
    magnitude = NullValue
    magnitudeError => Allocate2dSingle(1, nIndep, 1, nFilt, "magnitudeError")
    magnitudeError = NullValue
     

    !******************************************************
    ! Add the standard magnitude and its error to the table
    ! 
    !******************************************************
    n = 0
    
    do i = 4, nFilt - 1! NB NO STANDARDS FOR "WHITE OR UV FILTERS"
        stdString1 = trim(filtid(i)) // " - standard magnitude computed from "
        n = 0               
        if (stdMagExist(i))  then
   

            do j = 1, noColours
                if (nCount(i, j) .gt. 0) then
                    n = n + 1
                    magnitude(:, n) = NullValue
                    magnitudeError(:, n) = NULLValue
      
              
               
                    do k = 1, nIndep
                        magnitude(k, n) = stdMagnitudeStore(k, i, j)
                        magnitudeError(k, n) = stdMagnitudeErrorStore(k, i, j)
                    end do

                 
                         
                    comment = trim(stdstring1)
                    
                    stdstring1 = trim(comment) // ", " // trim(stdstring(j))                 
                    
     

                    call AddReal32IndexColumn(tableSet, trim(filtid(i)) // "_STD(" // trim(stdstring(j)) //")", &
                        "mag", "Standard magnitudes", nIndep, magnitude(:, n), RASortIndex, nullValue)

                    call AddReal32IndexColumn(tableSet, trim(filtid(i)) // "_STD_ERR(" // trim(stdstring(j)) // ")", &
                        "mag", "Standard magnitudes", nIndep, magnitudeError(:, n), RASortIndex, nullValue)

                  
                end if


            end do
        end if

        if (n .gt. 0) then

     
              
             call addcomment(tableset, trim(stdstring1))
     
               
            
        end if 

    end do
 
    string = "All filters"

    !********************************************
    ! Add the standard colours to the table
    !********************************************
    string = "All filters"
    ! Add the standard mags to the table



    do i = 1, noColours

        call AddReal32IndexColumn(tableSet, trim( colourString(i) ) // "_STD_COL", "mag", " ", &
                  nIndep, stdColourMagnitude(:, i), RASortIndex, nullValue)

        call AddReal32IndexColumn(tableSet, trim( colourString(i) ) // "_STD_COL_ERR", "mag", " ", &
                  nIndep, stdColourMagnitudeError(:, i), RASortIndex, nullValue)

    end do




 

    deallocate(magColour, magColourError, standardMagnitude, &
               standardMagnitudeError, stdColourMagnitude,   &
               stdColourMagnitudeerror, stat=allocErr)
    call CheckDeallocation(allocErr, "stdColourMagnitude")
    

  
    !**********************************************************
    ! Add the Vega Fluxes
    ! Use the flux conversion constants to compute fluxes
    !***********************************************************
!    call AddVegaFluxes(tableset, corrRate_table, corrRateErr_table, nIndep, rasortIndex, nofilters, filterindex, filtid)



    !**********************************************************
    ! Add the AB Fluxes
    ! Use the flux conversion constants to compute fluxes
    !***********************************************************
    call AddABMagsandFluxes(tableset, corrRate_table, corrRateErr_table, nIndep, rasortIndex, &
         nofilters, filterindex, filtid )





    !**********************************************************
    ! Add the White dwarf Fluxes
    ! Use the flux conversion constants to compute fluxes
    !***********************************************************
!    call AddWhiteDwarfFluxes(tableset, corrRate_table, corrRateErr_table, nIndep, rasortIndex, nofilters, filterindex, filtid)

    !**********************************************************
    ! Add the Agn Fluxes
    ! Use the flux conversion constants to compute fluxes
    !***********************************************************
!    call AddAgnFluxes(tableset, corrRate_table, corrRateErr_table, nIndep, rasortIndex, nofilters, filterindex, filtid)



 
   
    !*************************************
    ! Add the major-axes for each filter 
    !*************************************
    do j = 1 , noFilters
 
       i = filterIndex(j)
       label = trim(filtid(i)) // "_MAJOR_AXIS"       
       call Add2Real32IndexColumn(tableSet, label, "arcsec", "Semi-major axis", &
                       nIndep, majorTable, i, RASortIndex, nullValue)
    end do



    !*********************************************
    ! Add the mean major-axes 
    !*********************************************
    call AddMeanValTableColumn(tableSet, majorTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "MEAN_MAJOR_AXIS", "arcsec", "Mean major-axis (" // trim(filterString) //")", nullValue)
    source%semia = majorTable(:, 1)


    deallocate(majorTable, stat=allocErr)
    call CheckDeallocation(allocErr, "majorTable")
 
    !*********************************************
    ! Add the minor-axes for each filter
    !*********************************************



   do j = 1 , noFilters
 
       i = filterIndex(j)
       label = trim(filtid(i)) // "_MINOR_AXIS"       
       call Add2Real32IndexColumn(tableSet, label, "arcsec", "Semi-minor axis", &
                       nIndep, minorTable, i, RASortIndex, nullValue)
    end do





    !*********************************************
    ! Add the mean minor-axes 
    !*********************************************
    call AddMeanValTableColumn(tableSet, minorTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "MEAN_MINOR_AXIS", "arcsec", "Mean minor-axis (" // trim(filterString) //")", nullValue)
    source%semia = minorTable(:, 1)


    deallocate(minorTable, stat=allocErr)
    call CheckDeallocation(allocErr, "minorTable")

    !*********************************************
    ! Add the position-angle for each filter
    !*********************************************



     do j = 1 , noFilters
 
       i = filterIndex(j)
       label = trim(filtid(i)) // "_POSANG"       
       call Add2Real32IndexColumn(tableSet, label, "degreees", "Position angle of semi-major axis", &
                       nIndep, paTable, i, RASortIndex, nullValue)
    end do



    !*********************************************
    ! Add the mean position-angle 
    !*********************************************
    call AddMeanValTableColumn(tableSet, paTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "MEAN_POSANG", "arcsec",  "Mean position-angle " // trim(filterString), nullValue)


 


    ! Add the flags to the table 
 

    
    call addcomment(tableset, "QFLAG- bits set (value refers to integer value, colour refers to region colour)")
 

    

    call addcomment(tableSet, "QFLAG- bits set (colour refers to region colour)")

  
    call addcomment(tableSet, "if QFLAG=0 then region colour will be either")
    call addcomment(tableSet, "green (point-source) or cyan (extended source)")

    do i = 1, getNoQflags()

        call addcomment(tableSet, trim(getQFlagText(i)) )

    end do





    call addcomment(tableSet, "CFLAG- bits set (colour refers to region colour")



    do i = 1, getNoCflags()

        call addcomment(tableSet, trim(getCFlagText(i)) )

    end do



    call addcomment(tableset, "EG Bad pixel + read-out streak, bits 0 and 1 set, integer value =3 colour=red")





    !***************************************************
    !
    ! Now add the source flags
    !
    !***************************************************



    !**********************************************************************************************
    !
    ! The quality flags for each filter
    !   
    !********************************************************************************************** 
    do i = 1, noFilters
               

        j = filterIndex(i)

        if( omDataMode .ne. "MOSAICED" ) then
            !***********************************************************************
            ! Read-out streaks
            !**********************************************************************
            call ReadOutStreaks(int(nIndep, int32), rawRate_table(:, j), source%xpos, &
                       eFlagTable(:, j), qFlagTable(:, j) )



            !***********************************************************************
            ! Sources near to a bright source
            !**********************************************************************
            call BrightSourceProximityFlags(nIndep, rawRate_table(:, j), &
                source%xpos, source%ypos, qflagTable(:, j) )
 


        !***********************************************************************
        ! Sources within a smoke ring
        !**********************************************************************
!        call SmokeRings(nIndep, rawRate_table(:, j), source%xpos, source%ypos, &
!                eFlagTable(:, j), qFlagTable(:, j) )
        end if


        call Add2Int16IndexColumn(tableSet, trim( filtid(j)) // "_QUALITY_FLAG", &
               "", "Quality flag using combined source-list", &
               nIndep, qFlagTable, j, RASortIndex, int(int16nullValue, int32))
 



    end do    

  
 

    !*******************************************************************
    ! Quality and confusion flags for combined source-list
    ! Set source to the maximum in the tables for each flag
    !*******************************************************************
!    where(qflagTable .eq. int16NullValue) qflagTable = -1 ! was -1
!    where(cflagTable .eq. int8NullValue) cflagTable = -1_int8 ! was -1
!    where(eflagTable .eq. int8NullValue) eflagTable = -1_int8 ! was -1





    do i = 1, nIndep
        do j = 1, noFilters
            k = filterIndex(j)
            if(qFlagTable(i, k) .eq. int16NullValue) qFlagTable(i, k) = -1
            if(cFlagTable(i, k) .eq. int8NullValue)  cFlagTable(i, k) = -1
            if(eFlagTable(i, k) .eq. int8NullValue)  eFlagTable(i, k) = -1
        end do


        k = filterIndex(1)  ! The filter

        source(i)%qualityFlag = qflagTable(i, k )
        source(i)%confusionFlag = cflagTable(i, k)
        source(i)%extensionFlag = eflagTable(i, k)

!if( source(i)%extensionFlag == 66_int8) then
!write(*, *) "int8nullvalue=", int8nullvalue
!call fatal("bug at line 1050")
!end if      
 

        do j = 2, noFilters
  
            k = filterIndex(j)

            if (qflagTable(i, k) .gt. source(i)%qualityFlag) source(i)%qualityFlag = qflagTable(i, k)
            if (cflagTable(i, k) .gt. source(i)%confusionFlag) source(i)%confusionFlag = cflagTable(i, k)
            if (eflagTable(i, k) .gt. source(i)%extensionFlag) source(i)%extensionFlag = eflagTable(i, k)
            

        end do
       
    end do









    !***********************************************************************
    !
    ! Set flags (not on source-lists from mosaiced sky images)
    !
    !************************************************************************    
    if( omDataMode .ne. "MOSAICED" ) then
   
        call ReadOutStreaks(int(nIndep, int32), source%rawCountRate, source%xpos, &
                  source%extensionFlag, source%qualityFlag)


        !***********************************************************************
        ! Sources within a smoke ring
        !**********************************************************************

        call SmokeRings(nIndep, source%rawCountRate, source%xpos, source%ypos, source%qualityFlag)



        !***********************************************************************
        ! Sources near to a bright source
        !**********************************************************************


        call BrightSourceProximityFlags(nIndep, source%rawCountRate, &
                source%xpos, source%ypos, source%qualityFlag )
 

        !***********************************************************************
        ! Sources on a star diffraction spike
        !*********************************************************************
        source%theta = source%theta - pascx - 90.0
        where(source%theta .lt. 0.0) source%theta = source%theta + 180.0


        call StarSpikes(nIndep, source, source%xpos, source%ypos)
 


        !***********************************************************************
        ! Sources on an edge
        !**********************************************************************
        call EdgeSources(nIndep, source)



        !***********************************************************************
        ! Sources on a corner
        !**********************************************************************
        call setCornerFlags(source%xpos, source%ypos, source, int(nIndep, int16))

    end if

    where(source(1 : nindep)%qualityFlag .ge. 0) 
        qFlagTable(1 : nindep, 1) = source(1 : nindep)%qualityFlag
    elsewhere
         qFlagTable(1 : nindep, 1) = int16nullvalue
    end where

 
!
    !**************************************************************************************************
    ! Quality flag from combined-source list
    !**************************************************************************************************
    call Add2Int16IndexColumn(tableSet, "QUALITY_FLAG_COMB", "", "Quality flag using combined source-list", &
               nIndep, qFlagTable, 1, RASortIndex, int(int16nullValue, int32))
 
 


    !**********************************************************************************
    ! Confusion flags for each filter
    ! 
    !**********************************************************************************

    where(cFlagTable .lt. 0) cflagTable = int8nullvalue 
    do i = 1, noFilters

        call Add2Int8IndexColumn(tableSet, trim (filtid(filterIndex(i) )) // "_CONFUSION_FLAG", "", "Source confusion flag", &
               nIndep, cFlagTable, filterIndex(i), RASortIndex, int(int8nullValue, int32))

    end do



    cFlagTable(:, 1) = cflagTable(:, filterIndex(1))
    where(cFlagTable .eq. int8NullValue) cFlagTable = int(-1, int8)
  
    do i = 1, nIndep         
        cFlagTable(i, 1) = cFlagTable(i, filterindex(1))
        do j = 2, noFilters    
            k = filterIndex(j)
            if (cFlagTable(i, k) .ge. 0) then
                if (cFlagTable(i, k) .gt. cFlagTable(i, 1)) cFlagTable(i, 1) = cFlagTable(i, k)
            end if
        end do
    end do 

    where(cFlagTable(:, 1) .lt. int(0, int8)) cFlagTable(:, 1) = int8NullValue

    cflagTable(1 : nindep, 1) = source(1 : nindep)%confusionFlag

    call Add2Int8IndexColumn(tableSet, "CONFUSION_FLAG_MAX", "", "Maximum source confusion flag", &
               nIndep, cFlagTable, 1, RASortIndex, int(int8nullValue, int32))




    !**********************************************************************************
    ! Extension flags for each filter
    ! 
    !**********************************************************************************



    where(eFlagTable .lt. 0_int8) eflagTable = int8nullvalue 

 
    do i = 1, noFilters

        call Add2Int8IndexColumn(tableSet, trim (filtid(filterIndex(i) )) // "_EXTENDED_FLAG", "", "Source extension flag", &
               nIndep, eFlagTable, filterIndex(i), RASortIndex, int(int8nullValue, int32))

    end do






    eFlagTable(:, 1) = eflagTable(:, filterIndex(1))
    where(eFlagTable .eq. int8NullValue) eFlagTable = int(-1, int8)
  

    do i = 1, nIndep         
        eFlagTable(i, 1) = eFlagTable(i, filterindex(1))
 
        do j = 2, noFilters    
            k = filterIndex(j)
            if (eFlagTable(i, k) .ge. 0_int8) then
                if (eFlagTable(i, k) .gt. eFlagTable(i, 1)) eFlagTable(i, 1) = eFlagTable(i, k)
            end if

        end do
    end do 

    where(eFlagTable .lt. int(0, int8)) eFlagTable = int8nullvalue


    call Add2Int8IndexColumn(tableSet, "EXTENDED_FLAG_MAX", "", "Maximum source extension flag", &
               nIndep, eFlagTable, 1, RASortIndex, int(int8nullValue, int32))

 

    deallocate(rawRate_Table, corrRate_Table, corrRateErr_table)
    nullify(rawRate_Table, corrRate_Table, corrRateErr_table)




    if(associated(rawBackgroundRate_table) ) then ! Only used when SWS files comes from a sky-image
        deallocate(rawBackgroundRate_table)
        nullify(rawBackgroundRate_table)

    end if



    deallocate(eflagTable,stat=allocErr)
    call CheckDeallocation(allocErr, "eflagTable")
  



 

    !****************************************************************
    !
    ! Astrometry corection- optional 
    ! 
    !****************************************************************
    if (booleanParameter("usecat")) then
       !write(*,*) "L1517 ------------- usecat nSources=", nSources
       !write(*,*) "L1522 size(source)=", size(source)
       ! 2013-05-14 vny allocating raCorr and decCorr near the subroutine 
       ! DoAstrometryCorrection
       nSourcesModified=size(source)
       allocate(raCorr(1 : nSourcesModified)) 
       allocate(decCorr(1 : nSourcesModified))
       allocate(poserrCorr(1 : nSourcesModified))

       raCorr=source%rightascension
       decCorr=source%declination
       poserrCorr=source%poserr
       !do i =1, 3
       !   write(*,*) "L1536 ra=", source(i)%rightascension, " raCorr=", raCorr(i)
       !enddo
!        call release(outdataSet)
        
!        outDataSet = DataSet(stringParameter("outset"), modify)

!        tableSet = table(outDataSet, "SRCLIST")
       !do i = 1, nIndep
       !   if (i .lt. 5) then
       !      write(*,*) "L1505 ra=", source(i)%rightascension * radiansToDegrees, &
       !           " dec=", source(i)%declination * radiansToDegrees, " i=", i, &
       !           " raSortIndex=", raSortIndex(i)
       !       write(*,*) "L1523 raCorr=", raCorr(i) * radiansToDegrees, &
       !           " decCorr=", decCorr(i) * radiansToDegrees, " i=", i, &
       !           " raSortIndex=", raSortIndex(i)            
       !   endif
       !end do
             !**********************************************************
             ! NB RAs (source%rightAsxcension) and Decs (source%declination) 
             ! have to be in radians
             ! If successful then source%rightAscension and source%declination
             ! will contain the corrected coordinates
             !
             !*********************************************************

!             call DoAstrometryCorrection(nIndep,             &
!                                  raSortIndex,               &
!                                  nFilt,                     &
!                                  source%rightascension,     &
!                                  source%declination,        &
!                                  mag_table,                 &
!                                  source%poserr,             &
!                                  raz, decz,                 &
!                                  tableSet,                  &
!                                  oswFile)
       !write(*,*) "L1572 raz=", raz, " decz=", decz

             call DoAstrometryCorrection(nIndep,             &
                                  raSortIndex,               &
                                  nFilt,                     &
                                  raCorr,                    &
                                  decCorr,                   &
                                  mag_table,                 &
                                  poserrCorr,                &
                                  raz, decz,                 &
                                  tableSet,                  &
                                  oswFile)
             !************************************************************
             ! 2013-05-15 allocating and deallocating raCorr and decCorr
             ! near the subroutine DoAstrometryCorrection
             !************************************************************
             deallocate(raCorr, decCorr)
             deallocate(poserrCorr)

            comment = "Computed from corrected RAs and Decs"

        else

             comment = "Computed from uncorrected RAs and Decs"

    end if
   


    do ii=1, nfilt

       if (ii .eq. 1) then
          if (naverMagLimit(ii) .gt. 0.0) then
             call setAttribute(tableSet, "MLIMW2", averMagLimit(ii), &
                  "Detection limit for the filter UVW2")
          endif
       endif
       if (ii .eq. 2) then
          if (naverMagLimit(ii) .gt. 0.0) then
             call setAttribute(tableSet, "MLIMM2", averMagLimit(ii), &
                  "Detection limit for the filter UVM2")
          endif
       endif
       if (ii .eq. 3) then
          if (naverMagLimit(ii) .gt. 0.0) then
             call setAttribute(tableSet, "MLIMW1", avermagLimit(ii), &
                  "Detection limit for the filter UVW1")
          endif
       endif
       if (ii .eq. 4) then
          if (naverMagLimit(ii) .gt. 0.0) then
             call setAttribute(tableSet, "MLIMU", avermagLimit(ii), &
                  "Detection limit for the filter U")
          endif
       endif
       if (ii .eq. 5) then
          if (naverMagLimit(ii) .gt. 0.0) then
             call setAttribute(tableSet, "MLIMB", avermagLimit(ii), &
                  "Detection limit for the filter B")
          endif
       endif
       if (ii .eq. 6) then
          if (naverMagLimit(ii) .gt. 0.0) then
             call setAttribute(tableSet, "MLIMV", avermagLimit(ii), &
                  "Detection limit for the filter V")
          endif
       endif
       if (ii .eq. 7) then
          if (naverMagLimit(ii) .gt. 0.0) then
             call setAttribute(tableSet, "MLIMWH", avermagLimit(ii), &
                  "Detection limit for the filter WHITE")
          endif
       endif

    enddo

    position = columnNumber(tableSet, "POSERR") + 1
 


 


    !*************************************************** 
    ! Compute the galactic longitude and latitude 
    ! from the averaged RA and dec
    !***************************************************

    galong => AllocateDouble(1, nSources, "galong")
    galat => AllocateDouble(1, nSources, "galat")


    do i = 1, nIndep
        call sla_EQGAL(source(i)%rightascension , &
                       source(i)%declination ,    &
                       galong(i), galat(i) )
    end do

 
    galong = galong * radiansToDegrees
    

    call AddReal64IndexColumn(tableSet, "LII", "degrees", &
                 trim(comment), nIndep, galong, RASortIndex, nullValue, position=position)
    deallocate(galong, stat = allocErr)
    call CheckDeallocation(allocErr, "galong")

    galat = galat * radiansToDegrees
    position = position + 1
    call AddReal64IndexColumn(tableSet, "BII", "degrees", &
             trim(comment), nIndep, galat, RASortIndex, nullValue, position=position)
    deallocate(galat, stat = allocErr)
    call CheckDeallocation(allocErr, "galat")

    call release(tableSet)
    call release(outDataSet)


    deallocate(mag_table,magerr_table, magnitude, &
               stdMagnitudeStore, stdMagnitudeErrorStore, &
               magnitudeError, stat = allocErr)

    deallocate(source, paTable, stat=allocErr)
    deallocate(oswFile, RASortIndex, &
               nrows, stat=allocErr)
    call CheckDeallocation(allocErr, "nrows")
    deallocate(maglimit)
    deallocate(filtIndex)

    !
    ! 2013-01-14
    ! 2013-05-15 moving up to near the suroutine DoAstrometryCorrection
    !deallocate(raCorr, decCorr)
    !deallocate(poserrCorr)
    !
end subroutine omsrclistcomb





!***********************************************************************************
!
! Add a source ID column to each of the osw source lists
! telling us which source went to which entry in the
! observation source list 
! Input parameters
!      nosw - number of SWS files
!      oswfile - array storing the SWS file names
!      source - array containing all the original sources
!************************************************************************************
subroutine AddSourceIds(nosw, oswFile, nrows, nindep, source, raSortIndex)


    integer, intent(in) :: nosw, nindep
    character(*), pointer :: oswFile(:)
    integer, pointer :: nrows(:)
    type(sourceParameters), pointer :: source(:) ! All the input sources
    integer, pointer :: raSortIndex(:)
   

    integer(int32), pointer :: sourceID(:) => null(), tempIndex(:) => null()
   
    integer :: i, j, k, n
    type(dataSetT) :: oswDataSet
    type(TableT) :: tableSet
  

 
    write(errstr, '(a)') "Adding source ids"
    call message(AppMsg, Verbose)
          
    sourceID => allocateInt32(1, maxval(nrows), "sourceId")


    tempIndex => allocateInt32(1, nindep, "tempindex")


    do i = 1, nindep

       tempIndex(raSortIndex(i) ) = i
    end do
 


    do i = 1 , nosw ! Loop through each input source-list

!        write(*, *) "Adding source ids to file ", trim(oswfile(i)), " size=", nrows(i)    
        oswDataSet = DataSet(trim(oswFile(i)), modify)
        tableSet = table(oswDataSet, "SRCLIST", modify)

        n = nrows(i)
        if(n .gt. 0) then        


            !*******************************************************************
            ! Loop through each source in SWS file i
            !*******************************************************************

            do j = 1, size(source)
 
                if(source(j)%imin .eq. i) then     ! This source came from SWS file i
                    k = source(j)%imax             ! This is the source number in the file
                  

                    sourceID(k) = tempIndex( source(j)%noSourcePixels )

                end if
          
         
            end do
        end if

        !************************************************************************
        ! Add the SRC_ID column to SWS file i
        !************************************************************************
        call AddInt32Column(tableSet, sourceId, nRows(i), &
               "SRC_ID", "", "Source number in observation source-list file")
      


        call release(tableSet)

        call release(oswDataSet)

 
    end do



    deallocate(sourceId, tempIndex)
    nullify(sourceId, tempIndex)


end subroutine  AddSourceIds







!****************************************************
!
! CalculateStandardColour - Given the two filters, two
! instrumental magnitudes and colours, calculates the 
! standard colours. If the instrumental colours are outside
! the validity range the values are set to nullValue.
!
!*****************************************************
subroutine CalculateStandardColour(filterID1, filterID2, &
           instrumentalMagnitude1, instrumentalMagnitude1Error, & 
           instrumentalMagnitude2, instrumentalMagnitude2Error, &
           standardMagnitude1, standardMagnitude1Error, & 
           standardMagnitude2, standardMagnitude2Error, &
           standardColour, standardColourError)
  
    integer(int32) :: filterID1, filterID2
    real(double), intent(in) :: instrumentalMagnitude1, instrumentalMagnitude1Error
    real(double), intent(in) :: instrumentalMagnitude2, instrumentalMagnitude2Error
            
    real(double), intent(out) :: standardMagnitude1, standardMagnitude2                     
    real(double), intent(out) :: standardColour
    real(double), intent(out) :: standardColourError
    real(double), intent(out) :: standardMagnitude1Error
    real(double), intent(out) :: standardMagnitude2Error  
    real(double)    :: magnitudeRange
    type(OmColorTransformatorT)     :: colourTrans
    type(IntervalT)                 :: colourValInt


    
    colourTrans = CAL_omGetColorTransformator(filterID1, filterID2)
    colourValInt = CAL_omColorTransValidityRange(colourTrans)  


    !*******************************************************
    ! Check that the instrumental magnitudes lie within
    ! the allowed range.
    !*******************************************************
    magnitudeRange = instrumentalMagnitude1 - instrumentalMagnitude2

    if (magnitudeRange .gt. colourValInt%lower .and. &
        magnitudeRange .lt. colourValInt%upper)  then    ! Range okay                       
       
        !****************************************************
        ! NB The first magnitude in the following call is
        ! transformed into the standard system
        !****************************************************
        call CAL_omColorTransform(filterId1, filterID2, &
                  instrumentalMagnitude1, instrumentalMagnitude2, &
                  standardColour,  standardMagnitude2, &
                  standardColourError, standardMagnitude2Error)


        standardMagnitude1Error = AddErrors(standardMagnitude2Error,  standardColourError) 

!***************************************************************************
! Temporary (18.2.2002) Because the cal routine does not yet calculate the 
! standard magnitude errors properly, just set the error equal to
! the instrumental magnitude error
!***************************************************************************  

        standardMagnitude1Error = instrumentalMagnitude1Error
        standardMagnitude2Error = instrumentalMagnitude2Error
        standardMagnitude1 = standardMagnitude2 + standardColour
 
        standardColourError = adderrors( standardMagnitude1Error, standardMagnitude2Error )
 
    end if 
                                           
end subroutine CalculateStandardColour

!******************************************
!
! subroutine CreateObsSourceList
! Creates the output observation source-list file
!
!******************************************
subroutine CreateObsSourceList(noOswFiles, oswFile, outDataSet, raPoint, decPoint, exposure)
    type(DataSetT), intent(out) :: outDataSet
    integer, intent(in) :: noOswFiles
    character(*), intent(in), dimension(1 : noOswFiles) :: oswFile
    real(double), intent(out) :: raPoint, decPoint
    real(single), intent(out) :: exposure
    character(40) :: startTimeString, endTimeString
    type(DataSetT) :: oswDataSet
    integer i, k
    character(40) :: object, observer, obsID, startObs, endObs

    real(double) :: raPoint1, raPoint2
    real(double) :: nPointing=0.0

    outDataSet = DataSet(stringParameter("outset"), create)
  
    call setAttribute(outDataSet, "HDUCLASS", "OGIP", "Format conforms to OGIP/GSFC conventions")
    call setAttribute(outDataSet, "HDUCLAS1", "SRCLIST", "File contains a source list")
    call setAttribute(outDataSet, "FILTER", "n/a", "Not applicable")
    call setAttribute(outDataSet, "CONTENT", "SRCLIST", "File content")

    !************************************************
    ! 21.2.2002 - Add the new attributes for SCR 84
    !************************************************
    call OAL_addCommonAttributes(outDataSet)


    !************************************************
    ! Find the first start time and the last end time
    !************************************************
    oswDataSet = DataSet(oswFile(1), read)

    startTimeString = stringAttribute(oswDataSet, "DATE-OBS")
    endTimeString   = stringAttribute(oswDataSet, "DATE-END")

    !write(*,*) "L1923 startTimeString=", startTimeString

    if (hasAttribute(oswDataSet, "OBS_ID")) then
        obsID   = stringAttribute(oswDataSet, "OBS_ID")
    else
        obsId = "00000"
    end if
    if (hasAttribute(oswDataSet, "RA_PNT")) then
       !write(*,*) "L1931 has attribute RA_PNT"
       raPoint = real64Attribute(oswDataSet, "RA_PNT")
       ! store the value of the first raPoint for checking the 0/360 degree
       ! line crossing in diferent files
       raPoint1=raPoint
       decPoint = real64Attribute(oswDataSet, "DEC_PNT")
       nPointing=1.0
       !write(*,*) "L1934 i=1 raPoint=", raPoint, " raPoint1=", raPoint1 
    else
        raPoint = 0.0_double
        decPoint = 0.0_double
    end if
    !write(*,*) "L1939 raPoint=", raPoint
    object   = trim(stringAttribute(oswDataSet, "OBJECT"))
    !write(*,*) "L1941 object=", object

    if (hasAttribute(oswDataSet, "PA")) then
        pascx = real32Attribute(oswDataSet, "PA")
    else
        pascx = 0.0

    end if
    if (hasAttribute(oswDataSet, "OBSERVER")) then
 
        observer   = trim(stringAttribute(oswDataSet, "OBSERVER"))
    else
        observer = "MAX"
    end if
    exposure   = real32Attribute(oswDataSet, "EXPOSURE")
      
    call release(oswDataSet)
 
    do i = 2, noOswFiles
       !write(*,*) "L1961 i=", i, " ------"
       oswDataSet = DataSet(oswFile(i), read)
       
       startObs = stringAttribute(oswDataSet, "DATE-OBS")
       endObs   = stringAttribute(oswDataSet, "DATE-END")
       
       if (trim(stringAttribute(oswDataSet, "OBJECT")) .ne. object) &
            call warning("Different objects", "Object names in source-list files different")
       
       !******************************************************
       ! Need the camera pointing direction to do the 
       ! astrometry correction
       !***************************************************** 
       if (hasAttribute(oswDataSet, "RA_PNT")) then
          !write(*,*) "L1974 has attribute RA_PNT"
          ! check the crossing of the 0-360 degree line
          raPoint2=real64Attribute(oswDataSet, "RA_PNT")
          if (raPoint2>(raPoint1+180.0)) then
             call warning("RA coordinate", "RA_PNT likely to be near 360 degrees")
             raPoint2=raPoint2-360.0
          endif
          if (raPoint2<(raPoint1-180.0)) then
             call warning("RA coordinate", "RA_PNT likely to be near the 0  degrees")
             raPoint2=raPoint2+360.0
          endif
          !raPoint = raPoint + real64Attribute(oswDataSet, "RA_PNT")
          raPoint=raPoint+raPoint2
          decPoint = decPoint + real64Attribute(oswDataSet, "DEC_PNT")
          nPointing=nPointing+1.0
          !write(*,*) "L1978 raPoint=", raPoint, " nPointing=", nPointing
          
          k = CompareDates(startTimeString, startObs)
          
          if (k .lt. 0) startTimeString = startObs
          
          k = CompareDates(endTimeString, endObs)
          
          if (k .gt. 0) endTimeString = endObs
       else
          !write(*,*) "L1987 no attribute RA_PNT"
          call warning("Attribute", "No RA_PNT or DEC_PNT attribute found")
       end if
       call release(oswDataSet)  
    end do

    !write(*,*) "L1991 raPoint=", raPoint, " noOswFiles=", noOswFiles
    
    if (nPointing .gt. 1.0) then
       !raPoint = raPoint / dble(noOswFiles)
       !decPoint = decPoint / dble(noOswFiles)
       raPoint = raPoint / nPointing
       decPoint = decPoint / nPointing
    endif

    !write(*,*) "L1996 raPoint=", raPoint

    call setAttribute(outDataSet, "OBS_ID", obsID, "Observation identifier")
    call setAttribute(outDataSet, "OBJECT", object, "Object")
    call setAttribute(outDataSet, "OBSERVER", observer, "Observer")
 
    call setAttribute(outDataSet, "DATE-OBS", startTimeString, "Start time of observations")
    call setAttribute(outDataSet, "DATE-END", endTimeString, "End time of observations")
    call setAttribute(outDataSet, "RA_PNT", raPoint, "degrees", "Averaged RA pointing direction")
    call setAttribute(outDataSet, "DEC_PNT", decPoint, "degrees", "Averaged Dec pointing direction")
       

end subroutine CreateObsSourceList

!*******************************************************
! AddErrors
!
!*******************************************************
real(kind=double) function AddErrors(error1, error2)
    real(double), intent(in) :: error1, error2
 
    AddErrors = dsqrt(error1 * error1  + error2 * error2) 

end function AddErrors

!******************************************************
! subroutine SetUpInstrumentalColourTable
! Initialises the array with the possible colours
! NB Filter codes are V=V, U=U, B=B, White=W, Grism2=G,
! UVW1=L, UVM2=M, UVW2=S, Grism1=H.
!******************************************************
subroutine SetUpInstrumentalColourTable(Colour)

    type(ColourT), intent(out), dimension(:) :: Colour
   
    !***************** 
    ! These are OPT-OPT
    !*****************
    Colour(1) = ColourT(4, 5, "U_B", "U-B", "ERR_U_B", "")
    Colour(2) = ColourT(5, 6, "B_V", "B-V", "ERR_B_V", "")
    Colour(3) = ColourT(4, 6, "U_V", "U-V", "ERR_U_V", "")

    !****************
    ! These are UV-UV
    !****************
    Colour(4) = ColourT(1, 3, "S_L", "S-L", "ERR_S_L", "")
    Colour(5) = ColourT(1, 2, "S_M", "S-M", "ERR_S_M", "")
    Colour(6) = ColourT(2, 3, "M_L", "M-L", "ERR_M_L", "")

    !*****************
    ! These are UV-OPT
    !*****************
    Colour(7)  = ColourT(1, 4, "S_U", "S-U", "ERR_S_U", "")
    Colour(8)  = ColourT(1, 5, "S_B", "S-B", "ERR_S_B", "")
    Colour(9)  = ColourT(1, 6, "S_V", "S-V", "ERR_S_W", "")
    Colour(10) = ColourT(2, 4, "M_U", "M-U", "ERR_M_U", "")
    Colour(11) = ColourT(2, 5, "M_B", "M-B", "ERR_M_B", "")
    Colour(12) = ColourT(2, 6, "M_V", "M-V", "ERR_M_V", "")
    Colour(13) = ColourT(3, 4, "L_U", "L-U", "ERR_L_U", "")
    Colour(14) = ColourT(3, 5, "L_B", "L-B", "ERR_L_B", "")
    Colour(15) = ColourT(3, 6, "L_V", "L-V", "ERR_L_V", "")

end subroutine SetUpInstrumentalColourTable
!******************************************************
! subroutine SetUpStandardColourTable
! Initialises the array with the possible colours
! Letters refer to letter codes for filters
! ie V=V, U=U, B=B, W=White, G=Grism2, X=UVW1,
!   Y=UVM2, Z=UVW2 & H=Grism1
!******************************************************
subroutine SetUpStandardColourTable(Colour)

    type(ColourT), intent(out), dimension(:) :: Colour
     
    Colour(1) = ColourT(4, 5, "U_B", "U-B", "ERR_U_B", "_U_B")
    Colour(2) = ColourT(5, 6, "B_V", "B-V", "ERR_B_V", "_B_V")
    Colour(3) = ColourT(4, 6, "U_V", "U-V", "ERR_U_V", "_U_V")
    Colour(4) = ColourT(3, 4, "L_U", "L-U", "ERR_L_U", "_L_U")
    Colour(5) = ColourT(3, 5, "L_B", "L-B", "ERR_L_B", "_L_B")
    Colour(6) = ColourT(3, 6, "L_V", "L-V", "ERR_L_V", "_L_V")
    Colour(7) = ColourT(1, 5, "S_B", "S-B", "ERR_S_B", "_S_B")
    Colour(8) = ColourT(1, 6, "S_V", "S-V", "ERR_S_V", "_S_V")


end subroutine SetUpStandardColourTable
 
!*******************************************************
! CompareDates - Compares the first date with the second.
! Returns -1 if the second is less than the first,
! 0 if they are equal and 1 if greater.
!*******************************************************
integer function CompareDates(dateString1, dateString2)
    character(*), intent(in) :: dateString1, dateString2
    integer :: time1(6), time2(6)        ! Stores the time from the date string
    character(1) char    ! Used for date and time extraction 
    integer i

    read(dateString1, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time1(1), char, time1(2), char, time1(3), char, time1(4), &
         char, time1(5), char, time1(6)

    read(dateString2, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time2(1), char, time2(2), char, time2(3), char, time2(4), &
         char, time2(5), char, time2(6)
  
    do i = 1, 6

        CompareDates = time2(i) - time1(i)
        if (CompareDates .ne. 0) return
 
    end do

    CompareDates = 0 ! Must be equal

end function CompareDates
!************************************************************
!
!
! CompareOffsets
!
!************************************************************
subroutine CompareOffsets(fileName)
    character(*) :: fileName(:)
   
    real(single), pointer, dimension(:) ::  deltaRa, deltaRaError, &
                          deltaDec, deltadecError, &
                          deltaTheta, deltaThetaError
    real(single) :: deltaRamean, deltaRastandardDeviation
    real(single) :: deltaDecmean, deltaDecstandardDeviation
    real(single) :: deltaThetamean, deltaThetastandardDeviation
    real(single) :: minRaOffset, maxRaOffset
    real(single) :: minDecOffset, maxDecOffset

    character(5), pointer :: filter(:)

    logical, pointer :: status(:) => null()
    integer :: i, j, k
    integer(int8) :: noFilters
    character(5), pointer :: filterList(:) => null()
    character(4), pointer :: exposureNo(:) => null()


    integer(int32), pointer :: sourceIndex(:)
    integer(int32)  :: noFiles
  
    noFiles = size(fileName)
    allocate(status(1 : noFiles))
    write(errstr, *) "Checking astrometric offsets in each source-list file"
    call message(AppMsg, Verbose)
      


    deltaRa => allocateSingle(1, noFiles, "deltaRa")
    deltaRaError => allocateSingle(1, noFiles, "deltaRa")
    deltaDec => allocateSingle(1, noFiles, "deltaRa")
    deltaDEcError => allocateSingle(1, noFiles, "deltaRa")
    deltaTheta => allocateSingle(1, noFiles, "deltaTheta")
    deltaThetaError => allocateSingle(1, noFiles, "deltaTheta")
    sourceIndex => allocateInt32(1, noFiles, "sourceIndex")

    allocate(filter(1 : noFiles))
    allocate(filterList(1 : noFiles))
    allocate(exposureNo(1 : noFiles))
    

    k = 0
    minRaOffset = 1e6
    maxRaOffset = -1e6
    minDecOffset = 1e6
    maxDecOffset = -1e6

    noFilters = 0
iLoop : do i = 1, noFiles
 
        call GetOffsets(trim(fileName(i)), deltaRa(i), deltaRaError(i), &
                        deltaDec(i), deltadecError(i), &
                        deltaTheta(i), deltaThetaError(i), filter(i), exposureNo(i), status(i))
        if (status(i)) then
            k = k  + 1
            deltaRa(k) = deltaRa(i)
            deltaRaError(k) = deltaRaError(i)
            deltaDec(k) = deltaDec(i)
            deltaDecError(k) = deltaDecError(i)
            deltaTheta(k) = deltaTheta(i)
            deltaThetaError(k) = deltaThetaError(i)
            minRaOffset = min(minRaOffset,  deltaRa(k))
            maxRaOffset = max(maxRaOffset,  deltaRa(k))
            minDecOffset = min(minDecOffset,  deltaDec(k))
            maxDecOffset = max(maxDecOffset,  deltaDec(k))
            filter(k) = filter(i) 
            exposureNo(k) = exposureNo(i)
            do j = 1, noFilters 
                if (trim(filter(i)) == trim(filterList(j))) cycle iLoop
            end do
            noFilters = noFilters + 1
            filterList(noFilters) = trim(filter(i))
                    
        end if 
       
    end do iLoop 
    do i = 1, k
       
        do j = 1, noFilters 
            if (trim(filter(i)) == trim(filterList(j))) then
                sourceIndex(i) = j
                exit
            end if
        end do 
    end do 
    if (k > 1) then


        call ComputeAverage(deltaRa, deltaRaError, k, deltaRamean, deltaRastandardDeviation)
        call ComputeAverage(deltaDec, deltaDecError, k, deltaDecmean, deltaDecstandardDeviation)
        call ComputeAverage(deltaTheta, deltaThetaError, k, deltaThetamean, deltaThetastandardDeviation)
  
       
        write(errstr, '(a, i3)') "Number of computed astrometric offsets=", k
        call message(AppMsg, Verbose)
    
        write(errstr, '(a, sp, f6.2, a, sp, f6.2)') " Mean RA offset=", deltaRamean, "+/-", deltaRastandardDeviation
        call message(AppMsg, Verbose)
        write(errstr, '(a, sp, f6.2, a, sp, f6.2)') "Mean dec offset=", deltaDecmean, "+/-", deltaDecstandardDeviation
        call message(AppMsg, Verbose)
        write(errstr, '(a, sp, f6.2, a, f6.2, a)') " Ra offset range=", minRaOffset, " to ",  maxRaOffset, " arcsec"
        call message(AppMsg, Verbose)
        write(errstr, '(a, sp, f6.2, a, f6.2, a)') "Dec offset range=", minDecOffset, " to ",  maxDecOffset, " arcsec"
        call message(AppMsg, Verbose)
  
    end if

    deallocate(deltaRa, deltaRaError, deltaDec, deltaDecError, deltaTheta, deltaThetaError) 
    deallocate(filter, filterList, exposureNo, sourceIndex, status)
end subroutine CompareOffsets    
!*****************************************************
!
! Get the astrometry offsets from the source-list file        
!
!*****************************************************
subroutine GetOffsets(sourceList, deltaRa, deltaRaError, &
                        deltaDec, deltadecError, &
                        deltaTheta, deltaThetaError, filter, exposureNo, status)
    character(*), intent(in) :: sourceList
    real(single), intent(out) ::  deltaRa, deltaRaError, &
                          deltaDec, deltadecError, &
                          deltaTheta, deltaThetaError
    character(*), intent(inout) :: filter, exposureNo
    character(20) :: text
    logical, intent(out) :: status
    type(dataSetT) :: dataSetH
    type(tableT)   :: sourceTable
      dataSetH = dataSet(sourcelist, read)

    filter = trim(stringAttribute(dataSetH, "FILTER")) !(11 : 13) 
          
    if( hasAttribute( dataSetH, "EXP_ID" )) then
        text = trim(stringAttribute(dataSetH, "EXP_ID")) !(11 : 13) 
        exposureNo = text(11 : 13)
    else
        exposureNo = "000"
 
    end if 
 
    sourceTable = table(dataSetH, "SRCLIST")
    if (.not. hasAttribute(sourceTable, "RAOFFSET")) then
        status = .false.
        call release(sourceTable) 
        call release(dataSetH)
 
        return
    end if 
    
    if (.not. booleanAttribute(sourceTable, "POSCOROK")) then
        status = .false.
        call release(sourceTable) 
        call release(dataSetH)
 
        return
    end if 
    

    deltaRa = real32Attribute(sourceTable, "RAOFFSET")
    deltaRaError = real32Attribute(sourceTable, "RAOFFERR")
    deltaDec = real32Attribute(sourceTable, "DEOFFSET")
    deltaDecError = real32Attribute(sourceTable, "DEOFFERR")
    deltaTheta = real32Attribute(sourceTable, "ROT_CORR")
    deltaThetaError = real32Attribute(sourceTable, "ROT_ERR")
    


    call release(sourceTable) 
    call release(dataSetH)
    status = .true.
   
end subroutine GetOffsets

!*********************************************************************************
!
! ComputeAverageMagnitude
! Calculates the weighted average and
! standard error
!
!**********************************************************************************
subroutine ComputeAverageMagnitude(magnitude, error, size, mean, standardError)
    integer, intent(in) :: size        
    real(single), intent(in), dimension(1 : size) :: magnitude, error    
    real(single), intent(out) :: mean, standardError
    real(double) :: weight
    real(double) :: sum1, sum2
   
 
    integer i
    real(double)      :: flux
    real(double)      :: fluxError
  
    if (size .eq. 1) then
        mean = magnitude(1)
        standardError = error(1)
        return
    end if


 

    sum1 = 0.0_double
    sum2 = 0.0_double
  
 
    do i = 1, size

        if (magnitude(i) .ge. 0.0 .and. magnitude(i) .lt. 100.0) then
          
            flux = dble(10.0 ** (-0.4 * magnitude(i)))                 
            fluxError = dble(flux * max(0.01, error(i) ))
            weight = 1.0_double / (fluxError * fluxError)
    
            sum1 = sum1 + weight
            sum2 = sum2 + flux * weight
   

        end if
    end do   
    
 
    if(sum1 > 0.0) then
        mean = sum2 / sum1  ! The mean


        !***********************************************************
        ! The uncertainty in the weighted mean
        !**********************************************************
    
        standardError = sqrt( 1.0_double / sum1) 
    
        standardError = max(0.01, 1.085735 * standardError / mean)
 
  
        mean = -2.5 * log10(mean)
    else
        mean = magnitude(1)
        standardError = error(1)
    end if



end subroutine ComputeAverageMagnitude


!*************************************************
!
! Subroutine GetMagnitudes
! Obtains a list of magnitudes from the filter
! with the most magnitudes
!
!*************************************************
subroutine GetMagnitudes(nFilt, nIndep, &
            insMagnitudeStore,          & 
            magnitude, nullValue)


    integer(int32), intent(in) :: nFilt, nIndep
    real(single), pointer       :: insMagnitudeStore(:, :)   
    real(single), pointer       :: magnitude(:)   
    real(single), intent(in)    :: nullValue              ! Null
    integer(int32) :: i, k, count, maxCount, index1

   
    maxCount = 0
    index1 = 1
   

    do i = 1 , nFilt

   
        count = 0 
        do k = 1, nIndep
            if (insMagnitudeStore(k, i) > 0.0 .and.  &
                insMagnitudeStore(k, i) < 30.0) count = count + 1
 
                      
        end do
        if (count > maxCount) then
            maxCount = count
            index1 = i
                       
        end if  
          
     
   
    end do
 
    magnitude = NullValue
   
      
    do i = 1, nIndep
        magnitude(i) = insMagnitudeStore(i, index1)
 
    end do

end subroutine GetMagnitudes
!**************************************************************************************
!
!
!**************************************************************************************


!subroutine CorrectSourceListFile(fileName, deltaRamean,  deltaRaStandardDeviation, &
!                                 deltaDecmean, deltaDecStandardDeviation, &           
!                                 deltaThetaMean, deltaThetaStandardDeviation)
     
!    character(*), intent(in) :: fileName
!    real(single), intent(in) :: deltaRamean,  deltaRaStandardDeviation, &
!                                 deltaDecmean, deltaDecStandardDeviation, &           
!                                 deltaThetaMean, deltaThetaStandardDeviation
!    real(double) :: raPoint, decPoint
!    integer(int32) :: nrows
!    real(double), pointer :: ra(:), dec(:)
!    real(double) :: coeffs(6)
!    type(DataSetT) :: dataSetH
!    type(tableT) :: tableSet

!    write(errstr, *) "Adding RA/dec corrected columns in file ", &
!                trim(fileName), " using the computed mean offsets"
!    call message(AppMsg, Verbose)
  
       
!    dataSetH = DataSet(fileName, read)
    
!    tableSet = table(dataSetH, 0)
     
!    raPoint = real64Attribute(dataSetH, "RA_PNT")
!    decPoint = real64Attribute(dataSetH, "DEC_PNT")
          

!    nrows = numberOfRows(tableSet)

!    ra => allocateDouble(1, nrows, "ra")
!    dec => allocateDouble(1, nrows, "dec")
  
!    call ReadReal64Column(tableSet, "RA", nrows, 0, ra)
!    call ReadReal64Column(tableSet, "DEC", nrows, 0, dec)

              


!    call release(tableSet)
!    call release(dataSetH)
!    coeffs(1) =  deltaRamean * degreesToRadians / 3600.0_double
!    coeffs(2) =  deltaDecmean * degreesToRadians / 3600.0_double
!    coeffs(3) = 1.0_double
!    coeffs(4) = 1.0_double
!    coeffs(5) = 0.0_double
!    coeffs(6) =  deltaThetamean * degreesToRadians
!    call CorrectSourcePositions(nrows, ra, dec, &
!                                coeffs, raPoint, decPoint)

    

!    call AddCorrectionColumns(fileName, nrows, ra, dec, &
!                              deltaRamean,  deltaRaStandardDeviation, &
!                              deltaDecmean, deltaDecStandardDeviation, &           
!                              deltaThetaMean, &
!                              deltaThetaStandardDeviation, 0, dble(0.0))
    
!    deallocate(ra, dec)
        
!end subroutine CorrectSourceListFile

!*******************************************
! Function MissionTime - Returns the number
! of seconds from the start of the mission
! to the observation time. Added by CB because
! the call to CAL_GetBoresiteMatrix needs
! the time in seconds from the launch.
!******************************************
real(double) function  MissionTime(primaryTable)
    type(DataSetT), intent(in)   :: primaryTable

    character(len = 19)  :: dateObs         
    character(len = 19)  :: dateEnd        
    real(kind = double)  :: JulianDate0, JulianDate1, JulianDate2
    character(len = 19)  :: missionReferenceTime = "1998-01-01T00:00:00"

    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the start of the observations.
    dateObs = stringAttribute(primaryTable,  "DATE-OBS")    
    JulianDate1 = ModifiedJulianDate(dateObs)

    ! Compute the Julian date of the end of the observations.
    dateEnd = stringAttribute(primaryTable,  "DATE-END")   
    JulianDate2 = ModifiedJulianDate(dateEnd)

    ! The required time is the difference between the average of the 
    ! start and end times and the mission reference time, in seconds.
    MissionTime = ((JulianDate1 + JulianDate2) / 2.0d0 - JulianDate0) * 86400.0d0

end function MissionTime



!**********************************************************************
!
! Add the flags to the table
!*********************************************************************
subroutine AddFlagsToTable(tableSet, noFilters, filterIndex, nIndep, &
         flagTable, tableName, raSortIndex, intNullValue)


    type(tableT), intent(inout) :: tableSet
    integer, intent(in) :: noFilters
    integer :: filterIndex(:)
    integer, intent(in) :: nIndep
    integer(int8),  pointer, dimension(:, :)   :: flagTable
    character(*), intent(in) :: tableName
    integer, pointer :: raSortIndex(:)
    integer, intent(in) :: intNullValue    
    integer :: i, j, k, index
    character(14), pointer :: flagCode(:) => null()

    allocate(flagCode(1 : nIndep))

    flagCode = "       "
    


    do i = 1, nIndep
        flagCode = "--------"
        do j = 1, noFilters
            index = filterIndex(j)
            k = (j - 1) * 2 + 1
            if (flagTable(i, index) .ne. intNullValue) then
              
                write(flagCode(i)(k : k + 1), '(i2)') fLagTable(i, index)
        
              
            end if
!write(*, *) i, ") ", flagTable(i, index), " code= ", flagCode(i)
        end do
    end do 

    call AddCharIndexColumn(tableSet, tableName, " ", " ", nIndep, RASortIndex, flagCode)



    deallocate(flagCode)
    nullify(flagCode)

end subroutine AddFlagsToTable


!********************************************************************************
!
!
!
!*******************************************************************************


subroutine AddInstrumentalMags(tableSet, noFilters, filterIndex, &
             nIndep, mag_table, magerr_table, RASortIndex, nullValue)
    type(TableT), intent(inout) :: tableSet
    integer, intent(in) :: noFilters
    integer, intent(in) :: filterIndex(:)
    integer, intent(in) :: nIndep
    real(single), pointer :: mag_table(:, :), magerr_table(:, :)
    integer(int32), pointer :: RAsortIndex(:)
    real(single), intent(in) :: nullValue
    



!    do j  = 1 , noFilters

!        i = filterIndex(j)


!        label = trim(filtid(i)) // "_INS"
!        call Add2Real32IndexColumn(tableSet, label, "magnitudes", "Instrumental magnitudes", &
!                           nIndep, mag_table, i, RASortIndex, nullValue)

!        label = "ERR_" // trim(filtid(i)) // "_INS"
!        call Add2Real32IndexColumn(tableSet, label, "magnitudes", "Instrumental", &
!               nIndep, magerr_table, i, RASortIndex, nullValue)



!    end do







    call AddVectorColumn(tableSet, mag_table, nIndep, noFilters, filterIndex, RASortIndex, &
            "INS_MAG", "mags", "Instrumental", nullValue)

    call AddVectorColumn(tableSet, magerr_table, nIndep, noFilters, filterIndex, RASortIndex, &
            "INS_MAG_ERR", "mags", "Instrumental", nullValue)





end subroutine AddInstrumentalMags


!***********************************************************************
! This subroutine tries to fit the template-spectra to each source
! The best-fit template is used to compute the fluxes for the spectrum
! If no fit is done then fluxes are computed using the 
! zero-magnitude flux conversion constants from the ccf
!
!**********************************************************************


!subroutine ComputeSourceFluxes(nRows, omFitData, raSortIndex, tableSet)
!    integer, intent(in) :: nRows    
!    type(sourceData), pointer:: omFitData(:)
!    integer :: raSortIndex(:)
!    type(TableT), intent(inout) :: tableSet
  
  
!    integer :: noFits
!    type(fittedData) :: fit                                                      

!    logical(boolean) :: fitted
!    real(double) :: rmsMinimum
!    integer(int16) :: templateNo, noOfDegreesOfFreedom
!    integer :: i
 

   
 
    
    ! Load the stellar templates
!    call LoadTemplateData(templateData)

!    open(unit=10, status="UNKNOWN", file="fittedspectra.dat")
!    write(10, '(a12, a10, a10, a5)') "Source no", & 
!           "Temp No", "Rms Res", "MU"

!    call pgbegin(0, "spectra.ps" // '/VC', 1, 1)
  
       
!    noFits = 0


!    fit%flux = 0.0
!    fit%fluxErr = 0.0

    !******************************************************************
    ! Loop through all the sources and try a fit a template spectrum
    !******************************************************************
!    do i = 1, nRows
 
!        omFitData(i)%flux(1 : 7) = nullValue
!        omFitData(i)%fluxErr(1 : 7) = nullValue
  
        ! Try to fit a spectrum

!        call FitSpectra(i, omFitData(i), templateData, fitted, fit)
!        if (fitted) then ! Successful fit to template spectrum
!            call updateLogFile(i, rmsMinimum, templateNo,  noOfDegreesOfFreedom)
            
            ! Plot the fit
!            call PlotFit(i, fit)


!            omFitData(i)%flux = fit%flux
!            omFitData(i)%fluxErr = fit%fluxErr
!            noFits = noFits + 1   
         
!        end if
       
!    end do

!    call pgclos
!    write(10, '(a, i5, a, i5, a)') "Total number of fitted spectra=", noFits, " (no of sources=", nRows, ")"
      
!    write(errstr, '(a, i5, a, i5, a)') "Total number of fitted spectra=", noFits, " (number of sources=", nRows, ")"
   
!    call message(AppMsg, Verbose)
!    write(10, *) errstr          
!    close(10)
   



    ! Add the new flux columns
!    call AddTemplateFluxColumns(nRows, omFitData, raSortIndex, tableSet )


!end subroutine ComputeSourceFluxes

!****************************************************************
! This subroutine corrects the astrometry for the RA/DEC offset
! between the OM RA/DEC coordinate axes and the USNO
! RA/DEC axes
!
!
!**************************************************************
subroutine DoAstrometryCorrection(nIndep,       &
                                  raSortIndex,  &
                                  nFilt,        &
                                  ra,           &
                                  dec,          &
                                  mag_table,    &
                                  poserr,       &
                                  raz, decz,    &
                                  tableSet,     &  
                                  oswFile)

    integer, intent(in) :: nIndep, raSortIndex(:), nfilt
    !real(double), dimension(:) :: ra, dec
    real(double), pointer, dimension(:) :: ra, dec
    real(single), pointer, dimension(:, :) :: mag_table
    !real(single) :: poserr(:)
    real(single), pointer :: poserr(:)
    real(double), intent(in)    :: raz, decz
    character(*), pointer       :: oswFile(:)             ! filenames
    type(dataSetT) :: dataSetH
    type(TableT), intent(inout) :: tableSet
    real(double), pointer       :: raCorr(:)=>Null()      ! Correctewd RA
    real(double), pointer       :: decCorr(:)=>Null()     ! Corrected ded
    real(single), pointer :: magnitude(:) => null()
    real(single), pointer :: magTemp(:) => null()    
    real(single) :: maxRADecErr = 1.0, maxRmsResidual = 1.5
    logical :: status = .false.
    logical :: okay = .false.

  
    type(CorrectionData) :: corrData    
    type(CatalogueData), pointer :: catData(:) => null()   

    integer :: i 

    magTemp => AllocateSingle(1, nIndep, "magTemp")
    
    magnitude => AllocateSingle(1, nIndep, "magnitude")
 
    magnitude = NullValue
 


    !*****************************************************************
    ! Get the magnitudes for the astrometry correction
    !*****************************************************************
    call GetMagnitudes(nFilt, nIndep, mag_table, magnitude, nullValue)

    raCorr => AllocateDouble(1, nIndep, "raCorr")
    decCorr => AllocateDouble(1, nIndep, "decCorr")


  
    magTemp => allocateSingle(1, nIndep, "magTemp")
  



    !**************************************************************
    ! NB ra & dec have to be in radians  and in RA sorted order          
    !***************************************************************

    !write(*,*) "L2709 --------------------- "

    do i = 1, nIndep
        raCorr(i)  = ra(raSortIndex(i))
        decCorr(i) = dec(raSortIndex(i))
        magTemp(i) = magnitude(raSortIndex(i))
     
        !if (i .lt. 5) then
        !   write(*,*) "L2717 raCorr=", raCorr(i) * radiansToDegrees, &
        !        " decCorr=", decCorr(i) * radiansToDegrees, " i=", i, &
        !        " raSortIndex=", raSortIndex(i) 
        !endif        
    end do

    deallocate(magnitude)
    nullify(magnitude)   

  

  
    dataSetH = DataSet( oswfile(1), read)

    !*************************************************************
    ! Set the maximum number allowed matches for each source
    !
    !*************************************************************
    call setMaxNumMatches(10_int16)      
    
    !******************************************************************************
    !
    !******************************************************************************
    !write(*,*) "L2809 raz=", raz, " decz=", decz
    call CorrectAstrometry(nIndep, raCorr, decCorr,               &
                           posErr, magTemp, raz, decz,  dataSetH, &
                           status, corrData, catData)     



    call release( dataSetH )
      
    deallocate(magTemp)
    nullify(magTemp)
 


    if (status) then  ! The astrometry correction was successful
      
        maxRADecErr = realParameter("maxradecerr")
        maxRmsResidual = realParameter("maxrmsres")


 
        !******************************************************
        ! Check if fitted parameters are okay 
        !*****************************************************
        if (fitOkay(corrData, maxRADecErr, maxRmsResidual)) then
            okay = .true.

            !********************************************
            ! Need unsorted list
            !********************************************
 



            do i = 1, nIndep
                ra( rasortIndex(i) )  = raCorr(i)
                dec(rasortIndex(i) ) = decCorr(i)
            end do
 
            !************************************************************
            ! Check the mosaiced sky-image files
            ! 
            !************************************************************
     !       call checkimagefiles(corrData%xShift, corrData%yShift, decz)


        end if
    
  
    end if



    !**********************************************************
    ! Convert to degrees
    !*********************************************************

    raCorr = raCorr * radiansToDegrees
    decCorr = decCorr * radiansToDegrees

   

    !*********************************************************************
    ! Add RA_CORR and DEC_CORR correction columns 
    ! If the
    !*********************************************************************

    call AddCorrectionColumns2(tableSet, &
                 nIndep, raCorr, decCorr, corrData)
  
   
    call setAttribute(tableSet, "POSCOROK", okay, &
                              "ASTROMETRY CORRECTION OKAY")
    call setAttribute(tableSet, "SYSERROR", 0.5, &
                              "Systematic catalogue position errors")


    !*********************************************************************
    ! Check the computed offsets in each source-list file
    !*********************************************************************
    call CompareOffsets(oswFile)


end subroutine DoAstrometryCorrection

!*******************************************************
!
!
! Loads the data from a SWS file
!
!
!*******************************************************
subroutine LoadSourceData(oswFile, source, nStart, nRows)
    character(*), intent(in) :: oswFile
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: nStart
    integer, intent(inout) :: nRows

    type(TableT) :: tableSet
    type(DataSetT) :: oswDataSet
    integer(int32) :: windowx0, windowy0, xbin, ybin, filterIdInt
    real(double) :: plateScale
    character(5) :: omFilter
    real(single) :: exposure
    integer(int16) :: filterNo


    oswDataSet = DataSet(oswFile, read)

!    omDataMode = trim( stringAttribute(oswDataSet, "DATAMODE") )

    tableSet = table(oswDataSet, 0)
    nrows = numberOfRows(tableSet) ! The number of rows in the table

    exposure =  real32Attribute(oswDataSet, "EXPOSURE") ! The exposure time (secs)
    if(exposure .lt. 10) then
         omDataMode = "MOSAICED"
    else

        omDataMode = "IMAGING"
    
    end if


    source(nStart + 1 : nStart + nrows)%exposure = exposure
   


    ! Load the X and Y image coordinates
    call ReadReal32Column(tableSet, "XPOS", nrows, nStart, source%xpos)
    call ReadReal32Column(tableSet, "YPOS", nrows, nStart, source%ypos)

    ! Convert the pixel coordinates to full-frame coordinates (0.5- 2048.5, 0.5-2048.5)

    omFilter = stringAttribute(oswDataSet, "FILTER")
    !  Get the filter Id integer value
    filterIdInt = filterIdToInt(omFilter)



    !*********************************************************************
    ! Record the filter of these sources
    !*********************************************************************

    select case(trim(omFilter))
        case("UVW2", "uvw2")
            filterNo = 1
        case("UVM2", "uvm2")
            filterNo = 2
        case("UVW1", "uvw1")
            filterNo = 3
        case("U", "u")
            filterNo = 4
        case("B", "b")
            filterNo = 5
        case("V", "v")
            filterNo = 6
        case("WHITE", "white")
            filterNo = 7
        case default    ! Unrecognised filter-
            call release(oswDataSet)
             
            call fatal("unknown OM filter!", trim(omFilter))
     end select
    
    
    source(nStart + 1 : nstart + nrows)%jmin = filterNo



    call CAL_setState(OM, filterId = filterIdInt)


    ! Get the Plate scale (arcsecs/pixel)
    call CAL_omGetPlateScale(0.55_double, plateScale)


    call GetWindowParameters(oswDataSet, windowX0, windowY0, xbin, ybin)


    !*************************************************************
    !
    ! January 2008 - only convert if NOT from a sky-image
    !
    !**************************************************************
    if ( omDataMode .ne. "MOSAICED" ) then 

        call ConvertCoordsToFullFrame(source%xpos, source%ypos, nStart + 1, nStart + nRows, oswDataSet)   

    end if

   
    ! Get the position error
    call ReadReal32Column(tableSet, "POSERR", nrows, nStart, source%poserr)





    !***********************************************************************
    !
    ! Convert position errors to arc secs (originally in pixels)
    !
    !********************************************************************** 
    source(nStart + 1 : nstart + nrows)%poserr = &
               real(dble(source(nstart + 1 : nstart + nrows)%poserr) * plateScale * dble(xbin))


    source(nStart + 1 : nstart + nrows)%poserr =  max(0.1, source(nStart + 1 : nstart + nrows)%poserr)

!if (any(source(nstart + 1: nstart+nrows)%poserr .lt. 0.000001)) then
!call fatal("bug at line 3049")
!end if

 
    ! Get the Right-Ascension and declination
    call ReadReal64Column(tableSet, "RA", nrows, nStart, source%rightAscension)
    call ReadReal64Column(tableSet, "DEC", nrows, nStart, source%declination)

    ! Get the source major and minor axes and position angle
    call ReadReal32Column(tableSet, "FWHM_MAJ", nrows, nStart, source%semia)
    call ReadReal32Column(tableSet, "FWHM_MIN", nrows, nStart, source%semib)
    call ReadReal32Column(tableSet, "PA", nrows, nStart, source%theta)


    where(source(nstart + 1 : nstart + nrows)%theta .gt. 180.0) &
                source(nstart + 1 : nstart + nrows)%theta = source(nstart + 1 : nstart + nrows)%theta - 180.0
    
    ! Get the source significance
    call ReadReal32Column(tableSet, "SIGNIFICANCE", nrows, nStart, source%significance)

    ! Get the instrumental magnitudes
    call ReadReal32Column(tableSet, "MAG", nrows, nStart, source%magnitude)
    call ReadReal32Column(tableSet, "MAGERR", nrows, nStart, source%magnitudeerr)


    !**********************************************************************************
    !
    ! Get the raw rates
    !
    !**********************************************************************************
    call ReadReal32Column(tableSet, "RATE", nrows, nStart, source%rawCountRate)


    call ReadReal32Column(tableSet, "RATE_ERR", nrows, nStart, source%rawCountRateErr)

    !*************************************************************
    !
    ! December 2008 If using source-lists from MOSAICED sky images
    ! then need the total raw-rates (aperture corrected)
    !
    !**************************************************************
    if ( trim(omDataMode) == "MOSAICED" ) then 
     
        tolerance = 4.0
        call ReadReal32Column(tableSet, "BACKGROUND_RATE", nrows, nStart, source%rawBackgroundCountRate)
  
        !*********************************************************************************
        ! Get the photometry radius (unbinned)
        !********************************************************************************
        call ReadReal32Column(tableSet, "APERTURE", nrows, nStart, source%apertureRadius)
 
        call ApertureCorrection(oswDataSet, nrows, source(nstart + 1 : nstart + nrows) )

    end if
   



    ! Get the rates for spectral-fitting and for checking for sources near read-out streaks

    call ReadReal32Column(tableSet, "CORR_RATE", nrows, nStart, source%corrCountRate)
    call ReadReal32Column(tableSet, "CORR_RATE_ERR", nrows, nStart, source%corrCountRateErr)

    call ReadReal32Column(tableSet, "CORR_BACKGROUND_RATE", nrows, nStart, source%corrBackgroundCountRate)

!    call ReadReal32Column(tableSet, "CORR_BACKGROUND_RATE_ERR", nrows, nStart, source%corrBackgroundCountRateErr)


    !****************************************************************        
    ! Get the source flags (quality, confusion and extension)
    !****************************************************************

    if(columnDataType( column(tableSet, "QFLAG") ) .eq. 2) then ! integer8
        call ReadInt8Column(tableSet, "QFLAG", nrows, nStart, source%extensionflag)
        source%qualityFlag = source%extensionFlag
    else
        call ReadInt16Column(tableSet, "QFLAG", nrows, nStart, source%qualityflag)

    end if

    call ReadInt8Column(tableSet, "CFLAG", nrows, nStart, source%confusionflag)
    call ReadInt8Column(tableSet, "EFLAG", nrows, nStart, source%extensionflag)
  
    call release(oswDataSet)
       
end subroutine LoadSourceData
!******************************************************************
!
!
!******************************************************************
subroutine pruneSources(source, iStart, iFinish, nRows)
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: iStart, iFinish
    integer, intent(inout) :: nRows
    real(double) :: rate
    integer :: i, k

    nRows = 0
    k = iStart - 1

    do i = iStart, iFinish
        rate = source(i)%rawCountRate + source(i)%rawBackgroundCountRate
        if(rate <= 5.0) then ! Only keep sources whose total raw-count rate is no greater than 5.0
            nRows = nRows + 1
            k = k + 1
            if(i .ne. k) then
                source(k) = source(i)
            end if
        end if

    end do
 
end subroutine pruneSources
!***************************************************
!
! Function SourceMatch
! Tries to match the source, "source" to a catalogue
!          source, "catSource"
!***************************************************
logical function SourceMatch1(fileIndex,       &
                           source, catSource,  &
                           matched, tolerance, &
                           matchedIndex)
 
    
    integer, intent(in) :: fileIndex

    type(ERCoords), intent(in) :: source
    type(ERCoords), pointer :: catSource(:)


    logical :: matched(:)
    real(single), intent(in) :: tolerance
    integer, intent(out) :: matchedIndex  
      
    real(double) :: distance
    real(double) ::  error1
    real(single) :: error
   
  
    real(single) :: minDistance        
  
    integer(int32) i, k

    if(source%matches%numMatches .eq. 0) then

        SourceMatch1 = .false.
        return
    end if   
   

 
    error1 = source%poserr
        
    matchedIndex = -1  ! Not matched anything yet
    minDistance = 1e8

    do k = 1, source%matches%numMatches
        i = source%matches%index(k)
       

        if(matched(i)) cycle  ! Already matched       
                
        if(catSource(i)%imin .ne. fileIndex) cycle

        error = min(3.0_double, nsigma * AddErrors(error1, dble(catSource(i)%posErr)) )
                     
        error = max(tolerance, error)
        distance = source%matches%radius(k)          

        if (distance .le. error) then
  
            if(distance .lt. minDistance) then
            
                matchedIndex = i
                minDistance = distance

            end if
                
        end if

    end do

  
    if(matchedIndex .gt. 0) then ! Found a match
        sourceMatch1 = .true.
!        write(*, *) "Matched source- index=", matchedIndex

    else
        sourceMatch1 = .false.
    end if 
 


end function SourceMatch1   

!**************************************************************
!
!*************************************************************
subroutine SortDataRAOrder(source, n)
    type(sourceParameters) :: source(:)
 
    integer, intent(in) :: n 
    type(sourceParameters), pointer :: tempSource(:) => null()
  
    integer(int32), pointer :: raSortIndex(:) => null()
    integer :: i
    

    rasortindex => allocateint32(1, n, "rasortindex")

    allocate(tempSource(1 : n)) ! For storing the combined source-list
  


    tempsource = source

    call IndexSort(tempsource%rightAscension, RASortIndex, n)
    tempsource = source
  
    do i = 1, n
          tempsource(i) = source(rasortindex(i))
        
 
    end do

    source = tempSource
   
    deallocate(rasortindex, tempSource)
    nullify(rasortindex, tempSource)


end subroutine SortDataRAOrder


!**********************************************************
! Add the Vega Fluxes
! Use the flux conversion constants to compute fluxes
!***********************************************************
subroutine AddVegaFluxes(srcListTable, counts, countserr, nrows, rasortIndex, nfilters, filterindex, filter)
    type(TableT) :: srcListTable
    real(single), pointer, dimension(:, :) :: counts, countserr
    integer, intent(in) :: nrows
    integer, pointer:: rasortindex(:) 
    integer, intent(in) :: nfilters, filterIndex(:)
    character(*) :: filter(:)
    real(single), pointer :: fluxTable(:, :) => null()
    real(single), pointer :: fluxErrTable(:, :) => null()
    integer :: i, j, k, position
    character(100) :: comment
    ! Flux conversions taken from sas watch-out pages 
    real(single), parameter :: VegafluxConvert(7) = (/4.31e-15, 1.78e-15, 4.74e-16, 1.60e-16, 1.20e-16, 2.19e-16, -1.0/)
    type(dataSetT) dataSetH

    dataSetH = parent(srclistTable)

    fluxTable => allocate2dsingle(1, nrows, 1, 7, "fluxtable")
    fluxErrTable => allocate2dsingle(1, nrows, 1, 7, "fluxerrtable")
    fluxTable = nullValue
    fluxErrTable = nullValue
  
    call addcomment(srclisttable, "VEGA_FLUX - Computed fluxes using Vega calibration for the following filters")
   


    do i = 1, nfilters
        write(comment, *) i, "=", trim(filter(filterIndex(i)))  
        call addcomment(srclisttable, trim(comment))
   

    end do

    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .le. 1000.0) then 
                if (j .eq. 7) then
                    fluxTable(i, k) = nullValue
                    fluxErrTable(i, k) = nullValue
                else
                    fluxTable(i, k) =  counts(i, j) * VegafluxConvert(j)            
                    fluxerrTable(i, k) = countsErr(i, j) *  VegafluxConvert(j)        
                end if  
            end if 

        end do
    end do  

    if (hasColumn(srclisttable, "STD_COLS_ERR")) then    
        position = columnNumber(srclisttable, "STD_COLS_ERR") + 1
    else
        position = numberOfColumns(srclisttable)
    end if
 


    call AddVectorColumn(tab=srcListTable, data=fluxtable, ndata=nrows, &
            ncols=nfilters, storeindex=RASortIndex, &
            heading="VEGA_FLUX", units="erg/s/cm**2/A",  &
            comment="Fluxes based on Vega flux calibration ", &
            nullvalue=nullValue, position=position)

    call AddVectorColumn(tab=srcListTable, data=fluxerrtable, ndata=nrows, &
             ncols=nfilters, storeindex=RASortIndex, &
            heading="VEGA_FLUX_ERR", units="erg/s/cm**2/A", &
            comment="Flux errors based on Vega flux calibration ", & 
            nullvalue=nullValue, position=position + 1)

    deallocate(fluxtable, fluxerrtable)
    nullify(fluxtable, fluxerrtable)

end subroutine AddVegaFluxes






!**********************************************************
! Add the AB magnitudes and Fluxes 
! Use the flux conversion constants to compute fluxes
!***********************************************************
subroutine AddABMagsandFluxes(srcListTable, counts, countserr, nrows, &
           rasortIndex, nfilters, filterindex, filter)
    type(TableT) :: srcListTable
    real(single), pointer, dimension(:, :) :: counts, countserr
    integer, intent(in) :: nrows
    integer, pointer:: rasortindex(:) 
    integer, intent(in) :: nfilters, filterIndex(:)
    character(*) :: filter(:)
  
    real(single), pointer :: fluxTable(:, :) => null()
    real(single), pointer :: fluxErrTable(:, :) => null()
    integer :: i, j, k, position
    character(100) :: comment

    !*********************************************************************
    ! The following constants are taken from Antonio's e-mail of 21.07.05
    !*********************************************************************
    real(double) :: ABfluxConvert(7) = &
                (/5.7000000000000000d-15, 2.2100000000000d-15, 4.8200000000000d-16, &
                 1.9400000000000d-16, 1.250000000000d-16, 2.50000000000000d-16, -1.0_double/)

    real(double)  :: zeropoint(7) = &
             (/16.57190000000_double, 17.41200000000_double, 18.56620000000_double, &
                 19.18900000000_double, 19.08090000000_double, 17.92300000000_double, -1.0_double/)


    real(double), parameter :: e2jy(7) = (/1.4981d11, 1.7787d11, 2.8227d11, 3.9445d11, 6.7500d11, 9.8283d11, -1.0_double/)

    type(dataSetT) dataSetH
    type(OmFluxConvF) :: cf

    dataSetH = parent(srclistTable)

    fluxTable => allocate2dsingle(1, nrows, 1, 7, "fluxtable")
    fluxErrTable => allocate2dsingle(1, nrows, 1, 7, "fluxerrtable")

    fluxTable = nullValue
    fluxErrTable = nullValue
  
!    call addcomment(srclisttable, &
!         "AB Magnitudes and fluxes (AB_MAG, AB_MAG_ERR, AB_FLUX, AB_FLUX_ERR), computed for the following filters")
   
    do i = 1, nfilters
        j = filterIndex(i)            
        if (filter(j) .eq. "WHITE") cycle

        !*********************************************************************
        ! Get the photometry conversion constants from CAL
        !*********************************************************************  
        cf = getFluxConstants(filter(j))
        zeroPoint(j) = cf%zpab
        abfluxconvert(j) = cf%cfab
        
        !write(*, *) "L3423 i=", i, "=", trim(filter(j)), " zero-point=", zeroPoint(j), &
        !            " flux conversion constant=", abfluxconvert(j)  


        !write(*, *) "L3427 From cal zero-point=", cf%zpab, " flux conversion constant=", cf%cfab



!        call addcomment(srclisttable, trim(comment))

        write(comment, *) trim(filter(j)), " AB magnitude zero-point"  

        call setAttribute(srcListTable, "ABM0" // trim(filter(j)), cf%zpab, trim(comment))
        call setAttribute(srcListTable, "ABF0" // trim(filter(j)), &
             cf%cfab, "AB flux conversion constant (erg/s/cm**2/A")

    end do

    !*************************************************************
    ! Add the AB magnitudes

    !************************************************************
    !write(*,*) "L3445 adding AB magnitudes"
    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (filter(j) .eq. "WHITE") cycle

            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .lt. 1000.0) then 
                if (j .eq. 7) then
                    fluxTable(i, k) = nullValue
                    fluxErrTable(i, k) = nullValue
                else
                    fluxTable(i, k) =  zeropoint(j) - 2.5 * log10(counts(i, j))            
                    fluxerrTable(i, k) = countsErr(i, j) *  ABfluxConvert(j) 
                    !***************************************************
                    ! 2013-12-03 removing the 0.01 threshold of the magnitude error 
                    !fluxerrTable(i, k) = max(0.01, 1.085735 * countsErr(i, j) / counts(i, j))
                    fluxerrTable(i, k) = 1.085735 * countsErr(i, j) / counts(i, j)
                    !***************************************************
                end if

            end if 

        end do
    end do  

    if (hasColumn(srclisttable, "STD_COLS_ERR")) then    
        position = columnNumber(srclisttable, "STD_COLS_ERR") + 1
    else
        position = numberOfColumns(srclisttable)
    end if
 
    do i = nfilters, 1, -1
        if (trim(filter(filterIndex(i))) .eq. "WHITE") cycle
        call AddReal32IndexColumn(srcListtable, trim(filter(filterIndex(i))) // "_AB_MAG_ERR", "mag", " ", nRows, &
                    fluxErrTable(:, i), RASortIndex, nullValue, position)


        call AddReal32IndexColumn(srcListtable, trim(filter(filterIndex(i))) // "_AB_MAG", "mag", " ", nRows, &
                    fluxTable(:, i), RASortIndex, nullValue, position)
 
  
     end do


    !*************************************************************
    ! Add the AB fluxes
    !************************************************************
    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .le. 1000.0) then 
                if (j .eq. 7) then
                    fluxTable(i, k) = nullValue
                    fluxErrTable(i, k) = nullValue
                else
                    fluxTable(i, k) =  counts(i, j) * ABfluxConvert(j)            
                    fluxerrTable(i, k) = countsErr(i, j) *  ABfluxConvert(j)        
                

                end if
            end if 

        end do
    end do  




    do i = nfilters, 1,  -1
        if (trim(filter(filterIndex(i))) .eq. "WHITE") cycle
  
        call AddReal32IndexColumn(srcListtable, trim(filter(filterIndex(i))) // "_AB_FLUX_ERR", "erg/s/cm**2/A", " ", nRows, &
                    fluxErrTable(:, i), RASortIndex, nullValue, position)
 
        call AddReal32IndexColumn(srcListtable, trim(filter(filterIndex(i))) // "_AB_FLUX", "erg/s/cm**2/A", " ", nRows, &
                    fluxTable(:, i), RASortIndex, nullValue, position)
 
    end do
    

    do i = 1, nfilters
        j = filterIndex(i)
        if (filter(j) .eq. "WHITE") cycle

        comment = "E2JY" // trim( filter(j)) 
        call setattribute(srcListTable, trim(comment), e2jy(j), trim(filter(j)) // " flux conversion erg/s/cm**2/A to Jy")  
   
    end do

  


    deallocate(fluxtable, fluxerrtable)
    nullify(fluxtable, fluxerrtable)

end subroutine AddABMagsandFluxes
!**********************************************************
! Add the White dwarf  Fluxes
! Use the flux conversion constants to compute fluxes
!***********************************************************
subroutine AddWhiteDwarfFluxes(srcListTable, counts, countserr, nrows, &
            rasortIndex, nfilters, filterindex, filter)
    type(TableT) :: srcListTable
    real(single), pointer, dimension(:, :) :: counts, countserr
    integer, intent(in) :: nrows
    integer, pointer:: rasortindex(:) 
    integer, intent(in) :: nfilters, filterIndex(:)
    character(*) :: filter(:)
    real(single), pointer :: fluxTable(:, :) => null()
    real(single), pointer :: fluxErrTable(:, :) => null()
    integer :: i, j, k, position
    character(100) :: comment



 !      (/'UVW2 ','UVM2 ','UVW1 ','U    ','B    ','V    ', 'WHITE'/) 
    ! The following values are taken from Antonio's e-mail of 13.09.05    
    real(double), parameter :: whitedwarffluxConvert(7) = (/5.71d-15, 2.20d-15, 4.76d-16, 1.94d-16, &
            1.29d-16, 2.49d-16, -1.0_double/)

    type(dataSetT) dataSetH

    dataSetH = parent(srclistTable)


    fluxTable => allocate2dsingle(1, nrows, 1, 7, "fluxtable")
    fluxErrTable => allocate2dsingle(1, nrows, 1, 7, "fluxerrtable")
    fluxTable = nullValue
    fluxErrTable = nullValue
  
    call addcomment(srclisttable, &
         "WHITEDWARF_FLUX - Computed fluxes using White Dwarf calibration for the following filters")
   

    do i = 1, nfilters
        write(comment, *) i, "=", trim(filter(filterIndex(i)))  
        call addcomment(srclisttable, trim(comment))
     

    end do

    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .le. 1000.0) then 
                if (j .eq. 7) then
                    fluxTable(i, k) = nullValue
                    fluxErrTable(i, k) = nullValue
                else
                    fluxTable(i, k) =  counts(i, j) * whitedwarffluxConvert(j)            
                    fluxerrTable(i, k) = countsErr(i, j) *  whitedwarffluxConvert(j)        
             

                 end if           
                  
            end if 

        end do
    end do  

    if (hasColumn(srclisttable, "STD_COLS_ERR")) then    
        position = columnNumber(srclisttable, "STD_COLS_ERR") + 1
    else
        position = numberOfColumns(srclisttable)
    end if
 


    call AddVectorColumn(tab=srcListTable, data=fluxtable, ndata=nrows, &
            ncols=nfilters, storeindex=RASortIndex, &
            heading="WHITEDWARF_FLUX", units="erg/s/cm**2/A",  &
            comment="Fluxes based on White dwarf  flux calibration ", &
            nullvalue=nullValue, position=position)

    call AddVectorColumn(tab=srcListTable, data=fluxerrtable, ndata=nrows, &
             ncols=nfilters, storeindex=RASortIndex, &
            heading="WHITEDWARF_FLUX_ERR", units="erg/s/cm**2/A", &
            comment="Flux errors based on White dwarf  flux calibration ", & 
            nullvalue=nullValue, position=position + 1)

    deallocate(fluxtable, fluxerrtable)
    nullify(fluxtable, fluxerrtable)

end subroutine AddWhiteDwarfFluxes


!**********************************************************
! Add the AGN Fluxes
! Use the flux conversion constants to compute fluxes
!***********************************************************
subroutine AddAGNFluxes(srcListTable, counts, countserr, nrows, rasortIndex, nfilters, filterindex, filter)
    type(TableT) :: srcListTable
    real(single), pointer, dimension(:, :) :: counts, countserr
    integer, intent(in) :: nrows
    integer, pointer:: rasortindex(:) 
    integer, intent(in) :: nfilters, filterIndex(:)
    character(*) :: filter(:)
    real(single), pointer :: fluxTable(:, :) => null()
    real(single), pointer :: fluxErrTable(:, :) => null()
    integer :: i, j, k, position
    character(100) :: comment
    real(single) :: agnfluxConvert(7) = (/4.871e-15, 2.103e-15, 4.585e-16, 1.889e-16, 1.541e-16, 2.385e-16, -1.0/)
    type(dataSetT) dataSetH

    dataSetH = parent(srclistTable)

    fluxTable => allocate2dsingle(1, nrows, 1, 7, "fluxtable")
    fluxErrTable => allocate2dsingle(1, nrows, 1, 7, "fluxerrtable")
    fluxTable = nullValue
    fluxErrTable = nullValue
  
    call addcomment(srclisttable, "AGN_FLUX - Computed fluxes using AGN calibration for the following filters")
 
    do i = 1, nfilters
        write(comment, *) i, "=", trim(filter(filterIndex(i)))  
        call addcomment(srclisttable, trim(comment))
   

    end do

    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .le. 1000.0) then 
                fluxTable(i, k) =  counts(i, j) * agnfluxConvert(j)            
                fluxerrTable(i, k) = countsErr(i, j) *  agnfluxConvert(j)        
                  
            end if 

        end do
    end do  

    if (hasColumn(srclisttable, "STD_COLS_ERR")) then    
        position = columnNumber(srclisttable, "STD_COLS_ERR") + 1
    else
        position = numberOfColumns(srclisttable)
    end if
 


    call AddVectorColumn(tab=srcListTable, data=fluxtable, ndata=nrows, &
            ncols=nfilters, storeindex=RASortIndex, &
            heading="AGN_FLUX", units="erg/s/cm**2/A",  &
            comment="Fluxes based on Vega flux calibration ", &
            nullvalue=nullValue, position=position)

    !******************************************************************
    ! Add the flux errors
    !*****************************************************************
    call AddVectorColumn(tab=srcListTable, data=fluxerrtable, ndata=nrows, &
             ncols=nfilters, storeindex=RASortIndex, &
            heading="AGN_FLUX_ERR", units="erg/s/cm**2/A", &
            comment="Flux errors based on Vega flux calibration ", & 
            nullvalue=nullValue, position=position + 1)

    deallocate(fluxtable, fluxerrtable)
    nullify(fluxtable, fluxerrtable)

end subroutine AddAgnFluxes


!*********************************************************************
!
! Matches the source-lists from the different input SWS files
! to produce a merged source-list
!
! sources(k) k=1 to n, is the total number of sources inputted from all the input SWS files
!
! source(k)%imin        ! Remember which file this source came from
! source(k)%imax        ! Remember the row number in the source list file 
! source(k)%number      ! Need to remember the original source number
!
! source(k)%noSourcePixels ! After matching to find merged sources
!                          ! contains the merged-source number
!
!  
!
!********************************************************************
subroutine CreateMergedSourceList(n, source, nfiles, fileName, filter, &
                  nrows, nfilt, filterId, nnew, newSource)
    integer, intent(in) :: n
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: nFiles
    character(*), pointer :: fileName(:)
    character(*), pointer :: filter(:) ! List of filter for each file
    integer :: nrows(1 : nfiles)
    integer, intent(in) :: nFilt 
    character(*) :: filterId(1 : nfilt)
   
    integer, intent(out) :: nnew
    type(sourceParameters), pointer :: newSource(:)
   
    integer, pointer :: index(:, :), sourceCount(:)=> null()
    integer(int32) :: i, j, k, l1, l2
    type(sourceParameters), pointer ::tempSource1(:) => null()
    
    integer :: nFiltSources(1 : nfilt), maxFilt, maxFiltIndex 
    integer                     :: allocErr
    character(256) :: name
!real(double) distance  
    write(errstr, '(a)') "Creating merged source-list"
    call message(AppMsg, Verbose)
    write(errstr, '(a)') "Input source-list files"
    call message(AppMsg, Verbose)
  
    write(errstr, '(a8, a31, a30, a6, a4, a6)') "       ", "             name              ",  &
                                                "    ", " size ",   "   ",  "filter"
    call message(AppMsg, Verbose)
  
    sourceCount => allocateint32(1, n, "sourcecount")  
    ! Stores the number of sources that constitute a merged source
    index => allocate2dint32(1, n, 1, nfiles, "index") 
    ! Gives the indices, for each merged source, of the sources 
                                                       ! that constitute it


    k = 0     

    do i = 1, nfiles
        name = trim(fileName(i))
        l1 = scan( name, '/', BACK=.true. ) + scan( name, '\', BACK=.true. ) + 1
       
        l2 = len( name )
        write(errstr, '(i4, a4, a46, a13, i6, a4, a6)') i, ")   ", &
                name(l1 : l2), "    ",                            &
                nrows(i), "    ", trim(filter(i)) 
        call message(AppMsg, Verbose)



        do j = 1, nrows(i)
            k = k + 1
            source(k)%imin = i    ! Remember which file this source came from
            source(k)%imax = j    ! Remember the row number in the source list file 
            source(k)%number = k  ! Need to remember the original source number
  
        end do

       
    end do


    !******************************************************
    ! Find the filter that has the most sources
    !******************************************************

    nFiltSources = 0
    maxFilt = 0
   
    do j = 1, nfilt

        k = 1
        do i = 1, nfiles
  
           if ( source(k)%jmin .eq. j)  nfiltSources(j) = nfiltSources(j) + nrows(i)
 
           k = k + nrows(i)
       end do
        if (nfiltSources(j) .gt. 0) then
            write(errstr, '(a, a5, a, i6, a)') "Filter ", trim(filterid(j)), " has ", nfiltSources(j), " sources" 
            call message(AppMsg, Verbose)
            if (nFiltSources(j) .gt. maxFilt) then
                maxFiltIndex = j
                maxFilt = nFiltSources(j)
            end if
        end if

    end do

    !*************************************************
    ! Sort the data into increasing RA order
    !
    !************************************************      
    call SortDataRAOrder(source, n)


    !***************************************************************************
    !
    ! Match the sources
    ! On output, nnew stores the number of merged sources
    !
    !***************************************************************************
    call FindIdenticalSources(n, source, nfiles, nnew, index, sourceCount)


    !**************************************************************************
    ! Need now to compute the averaged parameters for the merged sources
    !**************************************************************************

    allocate(newSource(1 : nnew))
    allocate(tempSource1(1 : maxval(sourceCount)))
    
  
    !***************************************
    ! Allocate memory for output arrays.
    !***************************************
    mag_table => Allocate2DSingle(1, nnew, 1, nFilt, "mag_table")
    magerr_table => Allocate2DSingle(1, nnew, 1, nFilt, "magerr_table")

    ! Count-rate tables (for spectral fitting and CCD streak quality-flag setting)
    corrRate_table => Allocate2DSingle(1, nnew, 1, nFilt, "corrrate_table")
    corrRatEerr_table => Allocate2DSingle(1, nnew, 1, nFilt, "corrrateerr_table")

    rawRate_table => Allocate2DSingle(1, nnew, 1, nFilt, "rawrate_table")


    !************************************************************************
    ! If SWS file has come from a sky-image then need the following table for 
    ! raw=background count-rates
    !
    !*************************************************************************
    if( omDataMode .eq. "MOSAICED" ) then
   
        rawBackgroundRate_table => Allocate2DSingle(1, nnew, 1, nFilt, "rawbackgroundrate_table")

        rawBackgroundRate_table = NullValue

    end if

    !*************************************************************************
    !
    ! Significances
    !
    !*************************************************************************
    significanceTable => Allocate2DSingle(1, nnew, 1, nFilt, "significanceTable")

    ! Ellipse paramaterisation tables
    majorTable => Allocate2DSingle(1, nnew, 1, nFilt, "majorTable")
    minorTable => Allocate2DSingle(1, nnew, 1, nFilt, "minorTable")
    paTable => Allocate2DSingle(1, nnew, 1, nFilt, "paTable")

    ! Quality flag tables
    qflagTable => Allocate2DInt16(1, nnew, 1, nFilt, "qflagTable")
    cflagTable => Allocate2DInt8(1, nnew, 1, nFilt, "cflagTable")
    eflagTable => Allocate2DInt8(1, nnew, 1, nFilt, "eflagTable")
   

    allocate(magexist(1 : nnew, 1 : nFilt), stat=allocErr)
    call CheckAllocation(allocErr, "magExist")

    mag_table = NullValue
    magerr_table = NullValue
    corrRate_table = NullValue
    rawRate_table = NullValue

 
    corrRateErr_table = NullValue
    significanceTable = NullValue
    majorTable = NullValue
    minorTable = NullValue
    paTable = NullValue

    qflagTable = -1 ! int16nullValue
    cflagTable = -1_int8 ! int8nullvalue
    eflagTable = -1_int8 ! int8nullvalue
  

    magexist = .false.

    !********************************************************
    ! Compute the paramters for each merged source from
    ! the counterparts for each source
    !*******************************************************

    do i = 1, nnew
        do j = 1, sourceCount(i)
            tempSource1(j) = source(index(i, j)) 
        end do
 
        !*****************************************************************************
        ! Create the counts, etc, for this source (for each filter)
        !*****************************************************************************   
     
        call MergeSources(i, nFilt, &
                              sourceCount(i), &
                              tempsource1, &
                              newSource(i))    
    end do  

    where(qflagTable .lt. 0) qflagTable = int16nullvalue
    where(cflagTable .lt. 0_int8) cflagTable = int8nullvalue
    where(eflagTable .lt. 0_int8) eflagTable = int8nullvalue


    if (associated(tempSource1)) then
        deallocate(tempSource1)
        nullify(tempSource1)
    end if
   
   
    if(associated(sourceCount)) then
        deallocate(sourceCount)
        nullify(sourceCount)
    end if
  
    if(associated(index)) then
        deallocate(index)
        nullify(index)
    end if

end subroutine CreateMergedSourceList
!********************************************************************
!
! Subroutine MergeSource
! n input sources, source(1 : n) are merged into 1
! Also need to average parameters for the different filters
!
!*******************************************************************
subroutine MergeSources(i, nFilt, n, source, mergedSource)
    integer, intent(in) :: i
    integer, intent(in) :: nFilt
    
    integer, intent(in) :: n
    type(sourceParameters) :: source(:)
    type(sourceParameters), intent(out):: mergedSource
    integer :: j, k, l, iflag, nmag, nmag1
    real(double) :: ra_tmp, dec_tmp, poserr_tmp, weightsum, weight
    real(double) :: xi, eta
    real(single), dimension(1 : n) :: xPosTemp, yPosTemp, xyPosErrTemp
    real(single), dimension(1 : n) :: magTemp, magErrTemp,  corrRateTemp, corrRateErrTemp
    real(single) :: standardDeviation

    real(double) :: significanceTemp, majorTemp, minorTemp, paTemp
!    real(double) :: sumrawbackgroundCounts, sumrawCounts
    real(double), dimension(1 : nfilt) :: rawCountRateSum, rawBackgroundCountRateSum
    integer :: nRawCountRates(1 : nFilt)
    real(single) :: maxSourceSignificance, globalMaxSourceSignificance
    integer(int8) :: eFlagMaxExtended, globaleFlagMaxExtended
    
    !*************************************** 
    !
    ! Check to see if only one counterpart for 
    ! merged source i
    !
    !***************************************

    if (n .eq. 1) then ! Only one source counterpart
        mergedSource = source(1)
     

        do j = 1 , nFilt        ! Find which filter it belongs to


            if( source(1)%jmin .eq. j) then           
 
                qflagTable(i, j) = source(1)%qualityFlag 
                cflagTable(i, j) = source(1)%confusionFlag
                eflagTable(i, j) = source(1)%extensionFlag

       
                mag_table(i, j) = source(1)%magnitude

                magerr_table(i, j) = source(1)%magnitudeErr
                  
                corrRate_table(i, j) = source(1)%corrCountRate  
                corrRateErr_table(i, j) = source(1)%corrCountRateErr  

                rawRate_table(i, j) = source(1)%rawCountRate  

                !*************************************************************
                !
                ! Only need the following when SWS file comes from a sky-image
                !
                !***************************************************************
                if( associated(rawBackgroundRate_table)) then

                    rawBackgroundRate_table(i, j) = source(1)%rawBackgroundCountRate  
       
                end if

                significanceTable(i, j) = source(1)%significance
                majorTable(i, j) = source(1)%semia          
                minorTable(i, j) = source(1)%semib                  
                paTable(i, j)   = source(1)%theta

                if( source(1)%significance .gt. 0.0) then
                    magexist(i, j) = .true.
                end if
                return ! Found the filter so can return
            end if
        end do
     

    end if ! Finished case of single source


    !********************************************************************
    ! More than one counterpart for merged source i
    !
    !
    ! Set each bit of the quality and confusion flags to the maximum set
    !
    !********************************************************************
    mergedSource%qualityFlag = maxval(source(1 : n)%qualityFlag)         
    mergedSource%confusionFlag = maxval(source(1 : n)%confusionFlag)

    ra_tmp = 0.0_double
    dec_tmp = 0.0_double
    poserr_tmp = 0.0_double
    weightSum = 0.0_double
    majorTemp = 0.0_double
    minorTemp = 0.0_double
    paTemp = 0.0_double
    significanceTemp = 0.0_double

!    sumrawbackgroundCounts = 0.0_double
!    sumrawCounts = 0.0_double
 
    !********************************************************
    !
    ! Set each bit of the extension flag to the minimum set  
    !
    !******************************************************** 
    mergedSource%extensionFlag = minval(source(1 : n)%extensionFlag)

    do k = 1,  n ! Average
        if (source(k)%poserr .gt. 0.0) then
            weight = 1.0_double / (source(k)%poserr * source(k)%poserr)
        else
            weight = 1.0_double
        end if    
 
        weightSum = weightSum + weight
        !*************************************************************
        ! Convert RAs and decs to tangent-plane coordinates- average
        ! the sum of these and then convert the average to RA and dec
        !*************************************************************

        call sla_ds2tp(source(k)%rightascension, source(k)%declination, &
                    raz, decz, xi, eta, iflag)

        ra_tmp = ra_tmp + xi * weight     
        dec_tmp = dec_tmp + eta * weight
        poserr_tmp = poserr_tmp + source(k)%poserr * source(k)%poserr
        xPosTemp(k) = source(k)%xpos        ! X-coordinate
        yPosTemp(k) = source(k)%ypos        ! Y-coordinate
        xyPosErrTemp(k) = source(k)%poserr          

        majorTemp = majorTemp + dble(source(k)%semiA)
        minorTemp = minorTemp + dble(source(k)%semiB)
        paTemp = paTemp + dble(source(k)%theta)
        significanceTemp = significanceTemp + dble(abs( source(k)%significance))

!        sumrawbackgroundCounts = sumrawBackgroundCounts + source(k)%rawBackgroundCountRate * source(k)%exposure        
!        sumrawCounts = sumrawCounts + source(k)%rawCountRate * source(k)%exposure 
     end do
    

    ! The mean x coordinate
    call ComputeAverage(xPosTemp, xyposErrTemp, n, mergedsource%xpos, standardDeviation)        
    ! The mean y coordinate
    call ComputeAverage(yPosTemp, xyPosErrTemp, n, mergedsource%ypos, standardDeviation)        


    mergedsource%poserr = sqrt(poserr_tmp) / dble(n)

    mergedSource%semia = majorTemp / dble(n)
    mergedSource%semib = minorTemp / dble(n)
    mergedSource%theta = paTemp / dble(n)


    !**********************************************************
    ! Compute the average RA and declination 
    !**********************************************************

    ra_tmp = ra_tmp / weightSum
    dec_tmp = dec_tmp / weightSum
    call sla_dtp2s(ra_tmp, dec_tmp, raz, decz, mergedsource%rightascension, mergedsource%declination)


    mergedSource%significance = significanceTemp / dble(n)
 
   
    !*****************************************************
    ! Go through and fill up the magnitude matrix
    !  
    ! loop over all the filters
    !******************************************************

    rawCountRateSum = 0.0_double
   
    mergedSource%rawCountRate = 0.0
    nRawCountRates = 0

    globalMaxSourceSignificance = 0.0
  
    do j = 1 , nFilt ! Loop through all the filters
       
        nmag = 0
        nmag1 = 0
              
        significanceTemp = 0.0
        minorTemp = 0.0
        majorTemp = 0.0
        paTemp = 0.0

        !*********************************
        ! loop over all identical sources
        !*********************************

        l = 0

        !*********************************************
        ! Set the CAl for this filter (%jmin)
        !********************************************* 

        call setCalState(int(j, int8))


        
        maxSourceSignificance = 0.0
        eFlagMaxExtended = int8NullValue
        
        do k = 1 , n ! Loop through all the counterparts

            if( source(k)%jmin .eq. j) then   ! The same filters   
              

                !**********************************************************
                ! Do not use sources with negative significances
                ! for computing averaged count-rates, etc
                !**********************************************************
                if(  .not. isNullValue(source(k)%significance) ) then                
                    nmag = nmag + 1      
                    magTemp(nmag) = source(k)%magnitude
                    magErrTemp(nmag) = source(k)%magnitudeErr
                    corrRateTemp(nmag) = source(k)%corrCountRate
                    corrRateErrTemp(nmag) = source(k)%corrCountRateErr
                    significanceTemp = significanceTemp + abs(source(k)%significance)
                                        
                end if

                rawCountRateSum(j) = rawCountRateSum(j) + dble(source(k)%rawCountRate)         
                rawBackgroundCountRateSum(j) = rawBackgroundCountRateSum(j) + &
                             dble(source(k)%rawBackgroundCountRate)         

                nRawCountRates(j) = nRawCountRates(j) + 1

                if (.not. isNullValue(source(k)%semia)) then
                    nmag1 = nmag1 + 1
                    majorTemp = majorTemp + source(k)%semiA
                    minorTemp = minorTemp + source(k)%semiB
                    paTemp = paTemp + source(k)%theta
                end if

                qflagTable(i, j) = max( qflagTable(i, j), source(k)%qualityFlag ) 
                cflagTable(i, j) = max( cflagTable(i, j), source(k)%confusionFlag )    
                 
                if(abs(source(k)%significance) .gt. maxSourceSignificance) then
                    
                    eFlagMaxExtended = source(k)%extensionFlag
                    maxSourceSignificance = source(k)%significance
                end if 

                if(abs(source(k)%significance) .gt. globalmaxSourceSignificance) then

                      
                    globaleFlagMaxExtended = source(k)%extensionFlag
                    globalmaxSourceSignificance = source(k)%significance
                end if 

    
            end if  ! The same filter

        end do !  The number of counterparts
   

        !*******************************************************************************
        ! Set the extension flag for this filter to that of the most significant source
        !********************************************************************************

        eflagTable(i, j) = eFlagMaxExtended
      
        !*************************************************************************************
        !
        ! Compute the average values of the corrected count rates and instrumental magnitudes
        !
        !**************************************************************************************
        if (nmag .gt. 0) then
                    
            !*************************************************************
            ! APRIL 24th AVERAGE WITHOUT WEIGHTING
            !*********************************************************** 
            corrRate_table(i, j) = computeMean(nmag, corrRateTemp)


            !**************************************************************
            ! Compute standard error
            !**************************************************************
            corrRateErr_table(i, j) = computeStandardError(nMag, corrRateErrTemp)


            call computeInstrumentalMagnitude(corrRate_table(i, j), corrRateErr_table(i, j), &
                        mag_table(i, j), magErr_table(i, j) )

            if( corrRate_table(i, j) .gt. 1000.0) then
                magExist(i, j) = .false.
                significanceTable(i, j) = -significanceTemp / real(nmag)          
            else
            
                significanceTable(i, j) = significanceTemp / real(nmag)          
                magExist(i, j) = .true.
            end if
        else
            magExist(i, j) = .false.        
        end if 

        !**************************************************************************
        !
        !
        !**************************************************************************

        if (nmag1 .gt. 0) then
            majorTable(i, j) = majorTemp / dble(nmag1)          
            minorTable(i, j) = minorTemp / dble(nmag1)        
            paTable(i, j)   = paTemp / dble(nmag1)

        end if
             
    end do ! Looping through the filters


    mergedSource%rawCountRate = 0.0

    mergedSource%rawBackgroundCountRate = 0.0


    mergedSource%exposure = source(1)%exposure

    mergedSource%extensionflag = globaleFlagMaxExtended

!    write(*, *) "Mergesources line 4090" 
    do j = 1, nfilt
        if (nRawCountRates(j) .eq. 0) cycle 

        ! The mean raw count rate 
        mergedSource%rawCountRate = max(mergedSource%rawCountRate, real(rawCountRateSum(j) / &
             dble( nRawCountRates(j))) )

        rawRate_table(i, j) = real(rawCountRateSum(j) / dble(nRawCountRates(j)) )  
        if(associated (rawBackgroundRate_table) ) then

            rawBackgroundRate_table(i, j) = real(rawBackgroundCountRateSum(j) / dble(nRawCountRates(j)) )  
        end if 
 !       write(*, *) j, ") rawratetable=",    rawRate_table(i, j), " background=", rawBackgroundRate_table(i, j)
    end do


end subroutine MergeSources
!**********************************************************************
! Finds counterparts of a particular unique source in
! all the individual source-list files
!
! Input parameters-
!     n the total number of sources from all the input SWS files
!     source - array storing the sources
!     nFiles - the number of input SWS files
!
! Output parameters
!     nnew - the number of merged sources
!     index - 2d array giving the indices of the sources that
!             constitute a merged (unique) source
!     sourceCount - array giving the number of sources that 
!                   constitute each merged source
!
!**************************************************************************
Subroutine FindIdenticalSources(n, source, nfiles, nnew, index, sourceCount)
    integer, intent(in) :: n
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: nFiles
    integer(int32), pointer :: index(:, :)
    integer(int32), pointer :: sourceCount(:)
    logical, pointer :: matched(:) => null()
       
    integer(int32) :: i, j, k, m
    integer(int16) :: l
    integer, pointer :: sortIndex(:) => null()   
   
    integer(int32) :: nnew, ncount   
    real(single) :: tol
    


    !***************************************************************
    ! Store equatorial cartesian coordinates here
    !***************************************************************
    type(ercoords), pointer :: sourceCoords(:) => null()
 
  
    !*********************************************************************************
    ! Some initialization
    !
    !*********************************************************************************


  
    

    allocate(matched(1 : n))        ! Set to true when star matched, so not used again
    matched = .false.               ! Not matched anything


    !*************************************************************************************
    !
    !
    !*************************************************************************************/

    allocate(sourceCoords(1 : n))

    do i = 1, n
        sourceCoords(i)%rightAscension = source(i)%rightAscension
        sourceCoords(i)%declination = source(i)%declination
        sourceCoords(i)%cosdec = cos( source(i)%declination )
        sourceCoords(i)%poserr = source(i)%poserr
        sourceCoords(i)%imin = source(i)%imin
        sourceCoords(i)%number = source(i)%number
        sourceCoords(i)%noSourcePixels = source(i)%noSourcePixels
        sourceCoords(i)%semia = source(i)%semia
        sourceCoords(i)%extensionFlag = source(i)%extensionFlag


        call computeRectangularCoordinates2( source(i)%rightAscension, &
                  source(i)%declination, sourceCoords(i)%x,            &
                  sourceCoords(i)%y, sourceCoords(i)%z)


        !************************************************************************
        ! Set the errors for source-matching
        !
        !************************************************************************
        if(source(i)%extensionFlag .eq. 0) then 
         
            if ( btest(source(i)%qualityFlag, 4) ) then
                !************************************************************************
                ! Source probably surrounded by mod-8 pattern - centroiding not so good
                !************************************************************************
                sourceCoords(i)%posErr = max(source(i)%poserr, 1.0) 
 
            end if

        else
            !******************************************************************************************
            ! Extended source
            !******************************************************************************************
            sourceCoords(i)%posErr = max( source(i)%poserr, max(1.0, min(3.0, 0.25 * source(i)%semia)) )
        end if 


    end do ! Finished doing initialization

    !***************************************************************
    ! For each source, count the number of sources from other
    ! files that lie within a distance tolerance
    ! (store in sourceCoors%numNeighbours
    !***************************************************************
    call computeNumberNeighbours(sourceCoords, 5.0, int(nfiles * 10, int16))


    write(errstr,'(a, i6)') "Matching sources- number of input sources=", n
    call message(AppMsg, Verbose)
 
    nnew = 0 ! The number of sources in the merged source-list
      
    matched = .false.
    sourceCount = 0
    index = 0
    

    !*****************************************************
    ! Match the sources in order of decreasing significance
    !****************************************************
    sortIndex => allocateInt32(1, n, "sortIndex")
    call IndexSort(source%significance, sortIndex, n)


    !****************************************************************************************
    ! Loop through each source trying to find its counterparts in other files
    !
    !***************************************************************************************

!write(*, *) "Line 3926 tolerance=", tolerance

!write(*, *) "Findidenticalsources line 4021- nfiles=", nfiles
    nCount = 0 
lloop : do l = nfiles, 1,  -1

!        write(*, *) l, ") nnew=", nnew
        do j = 1, 5  ! was 4
            tol = 0.2 * real(j) * tolerance


            do m = n, 1, -1 ! n is the number of sources- loop through in order of decreasing significance
                i = sortIndex(m)                               ! Matching in order of decreasing significance
                if (matched(i)) cycle ! Has been assigned a counterpart to a new source, so skip

     
             
                nnew = nnew + 1                                ! The number of new sources

!                write(*, *) i, ") Matching source- ra=", sourceCoords(i)%rightAscension * radiansToDegrees, &
!                                                " dec=", sourceCoords(i)%declination * radiansToDegrees, &
!                           " nnew=", nnew


                sourceCoords(i)%noSourcePixels = nnew          ! Recall which merged source this source belongs to  

                sourceCount(nnew) = 1                          ! Only one match so far (itself) 
                index(nnew, 1) = i                             ! Record the index of the first source for this counterpart
         

                matched(i) = .true.                            ! Don't try to match source i with itself


                !*************************************************** 
                ! i is the current index in the source-list
                ! n is the final index
                ! nnew is the current new source  
                ! tol is the tolerance for matching
                !**************************************************
                call FindNeighbours(i, nnew, tol, nfiles,      &
                               sourceCount, index, sourceCoords, matched)



                !*****************************************************************
                ! If this source has not been matched with any other sources then
                ! remove this new source
                ! NB when l = 1 then probably only one source for this merged source
                !******************************************************************
                if(sourceCount(nnew) .lt. min(l, sourceCoords(i)%matches%numMatches + 1_int16) ) then  ! Remember l goes from 1 to the number of files
                 
                                  
                    do k = 1, sourceCount(nnew)
                        matched(index(nnew, k)) = .false.                     
                                                  
                    end do
                    nnew = nnew - 1 ! Discard this merged source for now
                else
                    ncount = ncount + sourceCount(nnew)
                end if 
               if (ncount .eq. n) exit lLoop ! Matched all the sources

         
            end do ! m loop - Looping through all the sources
        end do     ! j loop - looping through the tolerances
    end do lLoop        ! l loop 


    !*****************************************
    ! For testing purposes
    !****************************************
!    if (ncount .ne. n) then
!        write(*, *) "ncount=", ncount, " nnew=", nnew, " n=", n
!        call fatal("Bug at line 4054")
!    end if


    !***********************************************************
    ! Transfer matching details back to original source-list 
    !
    !***********************************************************   
    do i = 1, n
        
        source(i)%number = sourceCoords(i)%number
        source(i)%noSourcePixels = sourceCoords(i)%noSourcePixels
!        write(*, *) i, ") imin=", source(i)%imin, " nosourcepixels=", source(i)%noSourcePixels
 
!       if(source(i)%noSourcePixels .eq. 0 .or. source(i)%imin .eq. 0 .or. source(i)%imax .eq. 0) then
!            call fatal("bug at line 4055")
!        end if 

    end do

    deallocate(sourceCoords)
    nullify(sourceCoords)




!********************************************************
! For testing purposes only
!
!*******************************************************
!call listisolatedSources(nnew, sourceCount, source, index)



    !*****************************************************************
    ! Finished finding all the unique sources
    !*****************************************************************

    deallocate(sortIndex)
    nullify(sortIndex)

  
!**********************************************************
!
! For testing purposes only
!
!**********************************************************  

!write(*, *) "FINISHED FINDING IDENTICAL SOURCES"
!write(*, *) "LINE MATCHED SOURECES"
!l = 0


!do i = 1, nnew

!    l = l + sourceCount(i)
!    write(*, *) "NEW SOURCE ", i, ") counterparts="
!    write(*, '(10("index=", i6, " new no=", i6) )' )  ( index(i, j), &
!      source( index(i, j) )%noSourcepixels, j = 1, sourceCount(i) )

!    if(sourceCount(i) .eq. 0) then

!        call fatal("bug at line 3932")
!   end if

 


!   do j = 1, nnew
!        if(i .eq. j) cycle
!        do k = 1, sourceCount(j)
!            do m = 1, sourcecount(i)

!                if(index(j, k) .eq. index(i, m)) then
!                    write(*, *) "BUG SAME source counterpart for new sources ", i, " and ", j 
!                    write(*, *) "i=", i, " j=", j, " k=", k, " m=", m
!                    write(*, *) "index(", j, ", ", k, ")=", index(j, k)
!                    call fatal("bug at line 3942")
!                end if

!            end do
!        end do
!    end do
!end do    
 



!if(l .ne. n) then
!write(*, *) "l=", l, " n=", n
!call fatal("bug at line 3944")
!end if

    write(errstr, '(a, i6)') "Number of sources in combined source-list=", nnew
    call message(AppMsg, Verbose)
 
    deallocate(matched)
    nullify(matched)


end Subroutine FindIdenticalSources

!************************************************************
!
! Given a source, i, find all its neighbours,
! starting at index istart and finishing at n, the number
! of sources. Stars with matched(j) = true are skipped
!
!************************************************************
recursive subroutine FindNeighbours(i, nnew,         &
                     tolerance, nfiles,              & 
                     sourceCount,                    &
                     index,                          &
                     tempSource, matched)
    integer, intent(in) :: i, nnew
    real(single), intent(in) :: tolerance
    integer, intent(in) :: nfiles
   
    integer(int32), pointer :: sourceCount(:), index(:, :)
    integer :: k


    type(ERcoords), pointer :: tempSource(:)

    logical, pointer :: matched(:)
  
    integer(int32) :: matchedIndex

  
    if (sourceCount(nnew) .eq. nfiles) return


    !*********************************************************
    ! Find all matches to source i
    !*********************************************************
    do k = 1, nfiles  
        if( k .eq. tempSource(i)%imin) cycle   ! Source from the same file- ignore
        if( SourceMatch1(                    &
            k,                               &
            tempSource(i),                   &
            tempsource,                      &
            matched,                         &
            tolerance,                       &
            matchedIndex)) then

            sourceCount(nnew) = sourceCount(nnew) + 1    
            tempsource( matchedIndex )%noSourcePixels = nnew 

            matched(matchedIndex) = .true.
            index(nnew, sourceCount(nnew)) = matchedIndex 
             
        end if   
    end do

end subroutine FindNeighbours

!***************************************************************************
!
!
!
!**************************************************************************
subroutine AddMeanValTableColumn(tab, data, nData, nCols, colIndex, storeIndex, &
      heading, units, comment, nullValue)

    type(tableT), intent(inout) :: tab
    real(single), pointer :: data(:, :)
    integer, intent(in) :: nData, nCols
    integer, intent(in) :: colIndex(:), storeIndex(:)
    character(*), intent(in) :: heading, units, comment
    real(single), intent(in) :: nullValue
    integer :: i, j, k, l
    real(double) :: sum
    integer, dimension(1) :: s

    s(1) = 1

    do i = 1, nData
        k = 0
        sum = 0.0_double
        do j = 1, nCols
            l = colIndex(j)
            if ( .not. isNullValue(data(i, l)) .and. & 
                 data(i, l) .ne. nullValue .and. data(i, l) .ge. 0.0) then
           
                sum = sum + data(i, l)
                k = k + 1
            end if
        end do
        if (k .gt. 0) then
            if (k .eq. 1) then
                data(i, 1) = sum
            else
                data(i, 1) = sum / dble(k)
            end if
        end if
    end do

    call AddReal32IndexColumn(tab, heading, units, comment, nData, data(:, 1), storeIndex, nullValue)


end subroutine AddMeanValTableColumn

!********************************************************
!
!
!
!
!********************************************************
!subroutine checkMags(mergedNo, filter, n, mag, magerr, meanMag, stdMag, meanRa, meanDec, source)
!    integer, intent(in) :: mergedNo
!    character(*) :: filter
!    integer, intent(in) :: n
!    real(single) ::  mag(:), magerr(:)
!    real(single), intent(in)  :: meanMag, stdMag
!    real(double), intent(in) :: meanRa, meanDec
!    type(sourceParameters) :: source(1 : n)

!    integer, save :: count = 0
!    integer :: i, j

!    do i = 1, n

!        if ( abs(mag(i) - meanmag) .gt. 0.1) then
!            count = count + 1
!            write(10, '(i5, a, i5, a, a, a, f8.4, a, f8.4, a, f6.2, a, f6.2)') count, &
!                  " mergedno=", mergedNo, " Filter=", trim(filter), &
!                  " Mean right ascension=", meanRa, " dec=", meanDec, &
!             " mean mag=", meanMag, " error=", stdmag 
!            do j = 1, n
 !               write(10, '(i2, a, f6.2, a, f6.2, a, i2, a, i2, a, i2, a, f8.2, a, f8.2, a, a, a, i4, a, f4.2)') j, & 
 !                   ") mag=", mag(j), " error=", magerr(j), " cflag=", source(j)%confusionFlag, &
 !               " qflag=", source(j)%qualityFlag, " extension=", source(j)%extensionFlag, &
 !               " x=", source(j)%xpos, " y=", source(j)%ypos, " file name=", trim(fileName(source(j)%imin)), &
 !               " source no=", source(j)%imax, " poserr=", source(j)%poserr

!            end do
!           write(10, '(a, /, /)') " "

!            return
!         end if
        

!    end do

!end subroutine checkMags



!**********************************************************************
!
!
!
!**********************************************************************

real(single) function distanceerror(x, y, source)

    real, intent(in) :: x, y
    type(sourceParameters), intent(in) :: source
    real(single) :: dx, dy, delta, xdash, ydash, theta1

    theta1 = source%theta * degreesToRadians

    dx = source%xpos - x
    dy = source%ypos - y
    if ( abs(dx) < 1e-6 .and. abs(dy) < 1e-6 ) then
        delta = -theta1
    else
        delta = atan2(dy, dx) - theta1
    end if

    xDash = cos(delta) * source%semia
    yDash = sin(delta) * source%semib 
    
    distanceError = max(max(1.0, source%poserr), &
                   min(5.0, 0.25 * sqrt(xDash * xDash + yDash * yDash)))
    
end function distanceError
!*************************************************************************************
! For a given OM filter, this function returns a structure containing :
! 1) The magnitude zer0-point
! 2) The conversion factor from count-rate to flux
!
!
!*************************************************************************************
type(OmFluxConvF) function getFluxConstants( filter )
    character(*), intent(in) :: filter

    call CAL_setState(OM, filterId = filterIdToInt(filter))

    getFluxConstants = cal_omgetfluxconvfactors()
    
end function getFluxConstants

!************************************************************
!
! NB The following two subroutines are for  testing purposes
!
!***********************************************************
subroutine listisolatedSources(nnew, sourceCount, source, index)
    integer, intent(inout) :: nnew
    integer, pointer :: sourceCount(:), index(:, :) 
    type(sourceParameters), pointer :: source(:)

    integer :: i, j, k, count, l, m
    logical, pointer :: checked(:) => null()
    real(double) minDistance, distance

    write(*, *) "Isolated sources"
    allocate(checked(1 : size(source)))
    checked = .false.
    count = 0


    do i = 1, nnew ! The number of new sources
        if(checked(i)) cycle
        l = index(i, 1)  ! The first source in the list of countereparts for this unique source
        checked(i) = .true.
                        
        if( sourceCount(i) .lt. 3) then  ! This source has only one neighbour
            minDistance = 1e6
            k = 1
            do j = 1, nnew
                
                if (i == j) cycle
                if(checked(j) ) cycle
                m = index(j, 1)               
                if(source(l)%imin .eq. source(m)%imin) cycle
                distance = &
                          ArcLength(source(l)%rightAscension, source(l)%declination, &
                                    source(m)%rightascension, source(m)%declination) 
                if(distance .lt. minDistance) then
                    minDistance = distance
                    k = j
                end if  

            end do 
            if(minDistance .le. 7.0) then
                checked(k) = .true.

                count = count + 1 
                write(*, '(i5, a, f10.6, a, f10.6, a, i1, a, f6.2, a, f6.2, a, f6.2, a, f6.2, a, i4)' ) i, &
                         ") ra=", source(l)%rightAscension * radiansToDegrees, &
                          " dec=", source(l)%declination * radiansToDegrees, &
                    " eFlag=", source(l)%extensionFlag, " semia=", source(l)%semia,           &
                    " semib=", source(l)%semib, " pa=", source(l)%theta, " mindistance=", minDistance, &
                    " neighbour index=", k 
                
            end if
        end if
    end do

    deallocate(checked)
   write(*, *) "Nnew=", nnew, " Number of new sources with a neighbour <=3 arcsecs=", count
end subroutine listisolatedSources


!**************************************************************
! 
!
!*************************************************************
subroutine checkSourceList(nSource, source, minDist)
    integer, intent(in) :: nSource
    type(sourceParameters), pointer :: source(:)
    real(single), intent(in) :: minDist

    logical, pointer :: checked(:) => null()
    real(double) :: distance
    integer :: i, j, k
   

    write(*, *) "Checking source-list"
    allocate(checked(1 : nSource))
    checked = .false.
    k = 0
    do i = 1, nSource
       
!        write(*, *) i, ") Ra=", source(i)%rightAscension * radiansToDegrees      

!        if(checked(i)) cycle
        checked(i) = .true.    
     
 
        do j = 1, nSource
            if(i .eq. j) cycle               
!            if(checked(j)) cycle 
            
            distance = &
                      ArcLength(source(i)%rightAscension, source(i)%declination, &
                                source(j)%rightascension, source(j)%declination) 
            if(distance .le. minDist) then

                write(*, '(i5, a, f10.6, a, f10.6, a, i1, a, f6.2, a, f6.2, a, f6.2, a, f6.2, a, i4)' ) i, &
                         ") ra=", source(i)%rightAscension * radiansToDegrees, &
                          " dec=", source(i)%declination * radiansToDegrees, &
                    " eFlag=", source(i)%extensionFlag, " semia=", source(i)%semia,           &
                    " semib=", source(i)%semib, " pa=", source(i)%theta, " istance=", distance, &
                    " neighbour index=", j 

                

                k = k + 1   
            end if  

        end do 
            
    end do

    write(*, *) "Number of sources with a neighbour <=", minDist, " arcsecs=", k, " nnew=", nSource



    deallocate(checked)

end subroutine checkSourceList




!*************************************************************************************
!
! NB masterSource RAs and Decs are in radians
!
!*************************************************************************************
subroutine CheckIndividualSourceLists( masterSource, index )

    type (SourceParameters), pointer, dimension(:) :: masterSource
    integer, pointer :: index(:)
    integer :: i, j, k, noFiles, l1, l2
    type(dataSetT) :: dataSetH
    type(TableT) :: tableH
    character(256) :: fileName
    character(100) :: name
    integer :: noRows, nTotal
    real(double) :: distance
    type (SourceParameters), pointer, dimension(:) :: source => null()
   
    real(double) :: meanSeparation, maxSeparation
    real(double) :: meanRASeparation, maxRASeparation
    real(double) :: meanDEcSeparation, maxDecSeparation
    real(double) :: diff1, diff2, globalRMs, weight, sumWeight
    real(double) :: globalSumWeight    

 
!return
    noFiles = parameterCount( "sourcelistsets")

    if(noFiles .eq. 1) return

    globalRms = 0.0_double
    globalSumWeight = 0.0_double
    nTotal = 0


    write(errstr, *) "Checking astrometry"
    call message(AppMsg, Verbose)



    do i = 1, noFiles ! Loop through each one of the input source-list files
        
        !***********************************************************
        ! Load the data for this file
        !***********************************************************

        fileName = trim( stringParameter( "sourcelistsets", i - 1 ))

        name = trim(fileName)


        dataSetH = dataSet( trim(fileName), read )
        tableH = table( dataSetH, "SRCLIST")
        noRows = numberOfRows(tableH)
        if(noRows .eq. 0) then
            call release( dataSetH )
            cycle
        end if 

        allocate( source (1 : noRows ))



        call readInt32Column(tableH, "SRC_ID", noRows, 0, source%number)

        !****************************************************************************
        ! Want to check 
        !****************************************************************************

        call readreal64Column(tableH, "RA", noRows, 0,  source%rightAscension)
        call readreal64Column(tableH, "DEC", noRows, 0, source%declination)

        call readreal32Column(tableH, "POSERR", noRows, 0, source%poserr)


        call readreal32Column(tableH, "SIGNIFICANCE", noRows, 0, source%significance)
        call readreal32Column(tableH, "CORR_RATE", noRows, 0, source%corrCountRate)



        call release( dataSetH )
        nTotal = nTotal + norows
      
        maxSeparation = 0.0_double
        meanSeparation = 0.0_double
    
        maxRASeparation = 0.0_double
        meanRASeparation = 0.0_double
        maxDECSeparation = 0.0_double
        meanDECSeparation = 0.0_double
    
        sumWeight = 0.0_double

  
   
        do j = 1, norows
            k = index(source(j)%number)
            
          
            distance = ArcLength( mastersource(k)%rightAscension,                   &
                                      masterSource(k)%declination,                  &
                                      source(j)%rightascension * degreesToRadians,  &
                                      source(j)%declination * degreesToRadians)


!            if(distance .gt. 10.0) then
!                write(*, *) k, ") Master source- ra=", masterSource(k)%rightAscension * radiansToDegrees, " dec=", &
!                                                  masterSource(k)%declination * radiansToDegrees
!                write(*, *) j,") Matched source- ra=", source(j)%rightAscension, " dec=", &
!                                                   source(j)%declination
!                write(*, *) "distance=", distance, " source%number=", source(j)%number, " k=", k
!                call fatal("Bug at line 4578")
!            end if

            weight = 1.0_double / max(1.0, source(j)%poserr)
            sumWeight = sumWeight + weight
            globalsumWeight = globalsumWeight + weight
 

            diff1 = (masterSource(k)%rightAscension * radiansToDegrees) - source(j)%rightAscension
            diff2 = (masterSource(k)%declination  * radiansToDegrees)- source(j)%declination
            if(diff1 .gt. 180.0) diff1 = diff1 - 360.0  
 
            meanSeparation = meanSeparation + weight * distance
            globalRms = globalRms + weight * distance * distance
            maxSeparation = max(maxSeparation, distance)

            meanRASeparation = meanRASeparation + weight * diff1

            maxRASeparation = max(maxRASeparation, abs(diff1))

            meanDECSeparation = meanDEcSeparation + weight * diff2

            maxDECSeparation = max(maxDEcSeparation, abs(diff2))
                

        end do
      

        meanSeparation = meanSeparation / sumWeight
        meanRASeparation = 3600.0_double * meanRASeparation / sumWeight
        meanDEcSeparation = 3600.0_double * meanDECSeparation / sumWeight



        deallocate(source)
        nullify(source)

        name = trim(fileName)
        l1 = scan( name, '/', BACK=.true. ) + scan( name, '\', BACK=.true. ) + 1
       
        l2 = len( name )
      



        write(errstr, '(a, a31, a, i5)' ) "Checking File ", name(l1 : l2), " number of sources=", norows

        call message(AppMsg, Verbose)


        write(errstr, '(a, f6.2, a, f6.2, a)') "Mean separation=", meanSeparation, &
                    " maximum separation=", maxSeparation, " arcsecs"

        call message(AppMsg, Verbose)


        write(errstr, '(a, f6.2, a, f6.2, a)') "Mean RA separation=", meanRASeparation, &
                    " maximum separation=", maxRASeparation * 3600.0_double, " arcsecs"

        call message(AppMsg, Verbose)



        write(errstr, '(a, f6.2, a, f6.2, a)') "Mean DEC separation=", meanDECSeparation, &
                    " maximum separation=", maxDECSeparation * 3600.0_double, " arcsecs"

        call message(AppMsg, Verbose)



    end do

    if(nTotal .gt. 0) then
        globalRms = sqrt(globalRms / globalSumWeight)

        write(errstr, '(a, i6, a, f6.2, a)') "RMS separation computed from all sources (", &
            nTotal, ")=", globalRms, " arcsecs"

        call message(AppMsg, Verbose)

     
    end if
  
 
end subroutine CheckIndividualSourceLists

!******************************************************************************
! Computes an instrumental magnitude and error from a count-rate
!
!*****************************************************************************
subroutine computeInstrumentalMagnitude(rate, rateErr, magnitude, magnitudeError)
    real(single), intent(in) :: rate, rateErr
    real(single), intent(out) :: magnitude, magnitudeError

    real(double) :: corrMag, corrMagerrMinus, corrMagerrPlus

    if(rate .gt. 1000.0) then
        magnitude = nullValue
        magnitudeError = nullValue
    else

        call CAL_omPhotoMagnitude(dble(rate), corrMag, -dble(rateErr), &
                                  dble(rateErr), corrMagErrMinus, corrMagErrPlus)

        magnitude = real(corrMag)

        magnitudeError = 0.4342945 * 2.5 * rateErr / rate
   
    end if
end subroutine computeInstrumentalMagnitude

!**********************************************************************
!
!
!**********************************************************************
real(single) function computeMean(n, x)
    integer, intent(in) :: n
    real(single) :: x(:)
    
    computeMean = sum( dble( x(1 : n) ) )/ dble(n)


end function computeMean

!**********************************************************************
!
!
!**********************************************************************
real(single) function computeStandardError(n, x)
    integer, intent(in) :: n
    real(single) :: x(:)
    
    computeStandardError = sqrt( sum(dble(x(1 : n) ** 2)) ) / dble(n) 


end function computeStandardError
!******************************************************************
!
!
!*****************************************************************
subroutine setCalState(filterNo)
    integer(int8), intent(in) :: filterNo
    integer(int32) :: number

    number = filterNo + 12
    call CAL_setState(OM, filterId=number)

end subroutine setCalState

!***********************************************************
!
!
!
!
!
!
!**********************************************************
subroutine ApertureCorrection(dataSetH, nsources, source)
    type(DataSetT), intent(in) :: dataSetH
    integer, intent(in) :: nSources
    type(SourceParameters)  :: source(:)
    integer :: i
    real(double) :: frameTime, ratio, ratio1
    real(single) :: countsFrameRatio, apertureRadius

    frameTime = real64Attribute( dataSetH, "FRAMTIME") / 1000.0          
 
    apertureRadius = 12.0

!write(*, *) "frame time=", frameTime


!write(*, *) "Checkoutput Aperture correction line 1105"

    do i = 1, nSources
        if (source(i)%extensionFlag .eq. 1 .or. isNullValue( source(i)%apertureRadius) ) cycle
       
        if (abs(source(i)%apertureRadius - apertureRadius) .lt. 0.01) cycle

        if(frameTime < 1e-6) then
            countsFrameratio = min(real(source(i)%rawCountRate * frameTime), 1.0)
            countsFrameRatio = max(0.0, countsFrameRatio)
        else
            countsFrameRatio = 0.1
        end if 
!write(*, *) i, " countrate=", source(i)%rawCountRate, " aperture=", source(i)%apertureRadius

        ratio = PSFCountsRatio(apertureRadius, source(i)%apertureRadius, &
               1024.5, 1024.5, countsFrameRatio)

        source(i)%rawCountRate = source(i)%rawCountRate * ratio
        source(i)%rawCountRateErr = source(i)%rawCountRateErr * ratio

!write(*, *) i, " countrate=", source(i)%rawCountRate, " aperture=", &
!source(i)%apertureRadius, " ratio=", ratio

        ratio1 = apertureRadius / source(i)%apertureRadius
        ratio1 = ratio1 * ratio1

        source(i)%rawBackgroundCountRate = source(i)%rawBackgroundCountRate * ratio1
        source(i)%rawBackgroundCountRateErr = source(i)%rawBackgroundCountRateErr * ratio1


!if(abs(source(i)%rawCountrate) > 1e6) then
!write(*, *) "source aperture radius=", source(i)%apertureRadius, " ratio=", ratio, " ratio1=", ratio1, &
!           " raw count-rate=", source(i)%rawCountRate, " eflag=", source(i)%extensionFlag
!call fatal("bug")
!end if

    end do

end subroutine ApertureCorrection

end module omsrclistcomb_mod









